import sys
from pathlib import Path

import bpy
import numpy as np


print(" - Starting (alpha) blender megascript - ")

#######################################################################
##% Activate necessary addons
bpy.ops.preferences.addon_enable(module="io_import_images_as_planes")
bpy.ops.preferences.addon_enable(module="rigify")

#######################################################################

def get_video_paths(path_to_video_folder: Path) -> list:
    """Search the folder for 'mp4' files (case insensitive) and return them as a list"""

    list_of_video_paths = list(Path(path_to_video_folder).glob("*.mp4")) + list(
        Path(path_to_video_folder).glob("*.MP4")
    )
    unique_list_of_video_paths = get_unique_list(list_of_video_paths)

    return unique_list_of_video_paths


def get_unique_list(list: list) -> list:
    """Return a list of the unique elements from input list"""
    unique_list = []
    [unique_list.append(clip) for clip in list if clip not in unique_list]

    return unique_list

try:
    ##% Session path
    # #get session path as command line argument
    argv = sys.argv
    argv = argv[argv.index("--") + 1 :]
    session_path = argv[0]

    blend_file_save_path = argv[1]
    if len(argv) > 2:
        good_clean_frame_number = int(argv[2])
    else:
        good_clean_frame_number = 0
except:
    print("we appear to be running from the Blender Scripting tab! Manually enter your `session_path` at line 23")
    session_path = Path(
        r"C:\Users\jonma\Dropbox\FreeMoCapProject\FreeMocap_Data\sesh_2022-05-07_17_15_05_pupil_wobble_juggle_0"
    )
    good_clean_frame_number = 3341

print(f"Using session path - {str(session_path)}")
session_path = Path(session_path)

path_to_mediapipe_npy = (
    session_path / "output_data" / "raw_data" / "mediapipe3dData_numFrames_numTrackedPoints_spatialXYZ.npy"
)
path_to_mediapipe_3d_reproj = (
    session_path / "output_data" / "raw_data" / "mediapipe3dData_numFrames_numTrackedPoints_reprojectionError.npy"
)

print(f"Loading mediapipe data from {path_to_mediapipe_npy}")
#######################################################################
# %% load mediapipe data
# %%
print("loading data")
# %%
print(f"Loading mediapipe data from {path_to_mediapipe_npy}")
mediapipe_skel_fr_mar_dim = np.load(str(path_to_mediapipe_npy))
print("mediapipe_skel_fr_mar_dim.shape", mediapipe_skel_fr_mar_dim.shape)
mediapipe_skel_fr_mar_dim = mediapipe_skel_fr_mar_dim / 1000  # convert to meters

print(f"Loading mediapipe reprojection error data from {path_to_mediapipe_3d_reproj}")
mediapipe_reprojection_error_fr_mar = np.load(str(path_to_mediapipe_3d_reproj))
print(f"mediapipe_reprojection_error_fr_mar.shape = {mediapipe_reprojection_error_fr_mar.shape}")

body_marker_range = np.arange(0, 33)
right_hand_marker_range = np.arange(33, 33 + 21)
left_hand_marker_range = np.arange(33 + 21, 33 + 21 + 21)
face_marker_range = np.arange(left_hand_marker_range[-1], mediapipe_skel_fr_mar_dim.shape[1])

body_skel_fr_mar_dim = mediapipe_skel_fr_mar_dim[
    :,
    body_marker_range,
    :,
]
right_hand_fr_mar_dim = mediapipe_skel_fr_mar_dim[
    :,
    right_hand_marker_range,
    :,
]
left_hand_fr_mar_dim = mediapipe_skel_fr_mar_dim[
    :,
    left_hand_marker_range,
    :,
]
face_fr_mar_dim = mediapipe_skel_fr_mar_dim[
    :,
    face_marker_range,
    :,
]

# %% Session Specific stuff
number_of_frames = mediapipe_skel_fr_mar_dim.shape[0]
start_frame = 1
end_frame = number_of_frames

bpy.context.scene.frame_start = start_frame
bpy.context.scene.frame_end = end_frame

########################
## Find Good Clean Frame with some fiddly nonsense
if good_clean_frame_number == 0:
    frame_nan_counts = []
    frame_mean_reproj_error = []

    for this_frame in range(mediapipe_skel_fr_mar_dim.shape[0]):
        frame_nan_counts.append(np.sum(np.isnan(mediapipe_skel_fr_mar_dim[this_frame, :, 0])))
        frame_mean_reproj_error.append(np.nanmean(mediapipe_reprojection_error_fr_mar[this_frame, :]))

    nan_times_vis = np.array(frame_nan_counts) * np.array(frame_mean_reproj_error)
    num_frames = len(frame_nan_counts)
    # nan_times_vis[0:int(num_frames/5)] = np.nanmax(nan_times_vis)
    # nan_times_vis[-int(num_frames/5):-1] = np.nanmax(nan_times_vis)

    good_clean_frame_number = np.nanargmin(
        nan_times_vis
    )  # the frame with the fewest nans (i.e. hopefully a frame where all tracked points are visible)

print(f"----good_clean_frame_number: {good_clean_frame_number}----")

#######################################################################
# %% Mediapipe tracked point names
mediapipe_tracked_point_names = [
    "nose",
    "left_eye_inner",
    "left_eye",
    "left_eye_outer",
    "right_eye_inner",
    "right_eye",
    "right_eye_outer",
    "left_ear",
    "right_ear",
    "mouth_left",
    "mouth_right",
    "left_shoulder",
    "right_shoulder",
    "left_elbow",
    "right_elbow",
    "left_wrist",
    "right_wrist",
    "left_pinky",
    "right_pinky",
    "left_index",
    "right_index",
    "left_thumb",
    "right_thumb",
    "left_hip",
    "right_hip",
    "left_knee",
    "right_knee",
    "left_ankle",
    "right_ankle",
    "left_heel",
    "right_heel",
    "left_foot_index",
    "right_foot_index",
    "right_hand_wrist",
    "right_hand_thumb_cmc",
    "right_hand_thumb_mcp",
    "right_hand_thumb_ip",
    "right_hand_thumb_tip",
    "right_hand_index_finger_mcp",
    "right_hand_index_finger_pip",
    "right_hand_index_finger_dip",
    "right_hand_index_finger_tip",
    "right_hand_middle_finger_mcp",
    "right_hand_middle_finger_pip",
    "right_hand_middle_finger_dip",
    "right_hand_middle_finger_tip",
    "right_hand_ring_finger_mcp",
    "right_hand_ring_finger_pip",
    "right_hand_ring_finger_dip",
    "right_hand_ring_finger_tip",
    "right_hand_pinky_finger_mcp",
    "right_hand_pinky_finger_pip",
    "right_hand_pinky_finger_dip",
    "right_hand_pinky_finger_tip",
    "left_hand_wrist",
    "left_hand_thumb_cmc",
    "left_hand_thumb_mcp",
    "left_hand_thumb_ip",
    "left_hand_thumb_tip",
    "left_hand_index_finger_mcp",
    "left_hand_index_finger_pip",
    "left_hand_index_finger_dip",
    "left_hand_index_finger_tip",
    "left_hand_middle_finger_mcp",
    "left_hand_middle_finger_pip",
    "left_hand_middle_finger_dip",
    "left_hand_middle_finger_tip",
    "left_hand_ring_finger_mcp",
    "left_hand_ring_finger_pip",
    "left_hand_ring_finger_dip",
    "left_hand_ring_finger_tip",
    "left_hand_pinky_finger_mcp",
    "left_hand_pinky_finger_pip",
    "left_hand_pinky_finger_dip",
    "left_hand_pinky_finger_tip",
]

#####################################################################
## Blender Stuff
#####################################################################

try:
    #####################################################################
    ###%% clear scene
    print("clearing scene")
    try:
        bpy.ops.object.mode_set(mode="OBJECT")
    except:
        pass

    try:
        bpy.ops.object.select_all(action="SELECT")  # select all objects
        bpy.ops.object.delete(use_global=True)  # delete all objects from all scenes
    except:
        pass

    #######################################################################
    # %% create world origin

    bpy.ops.object.empty_add(type="ARROWS")
    world_origin_axes = bpy.context.active_object
    world_origin_axes.name = "world_origin"  # will stay at origin
    # inspect(world_origin_axes, methods=True)

    bpy.ops.object.empty_add(type="ARROWS")
    freemocap_origin_axes = bpy.context.active_object
    freemocap_origin_axes.name = (
        "freemocap_origin_axes"  # will translate to put skelly on ground symmetric-ish about origin
    )

    #######################################################################
    # %% load empties
    print("loading {} empties on {} frames".format(len(mediapipe_tracked_point_names), number_of_frames))

    empty_size = 0.01

    for this_point_index, this_point_name in enumerate(mediapipe_tracked_point_names):
        print(f"loading {this_point_name}...")
        bpy.ops.object.empty_add(type="SPHERE")
        this_empty = bpy.context.active_object
        this_empty.name = this_point_name

        if "face" in this_point_name:
            this_empty.scale = [empty_size / 4] * 3
        if "hand" in this_point_name:
            this_empty.scale = [empty_size / 2] * 3
        else:
            this_empty.scale = [empty_size] * 3

        this_empty.parent = freemocap_origin_axes

        this_point_fr_xyz = mediapipe_skel_fr_mar_dim[:, this_point_index, :]

        for frame_num in range(end_frame):
            this_empty.location = [
                this_point_fr_xyz[frame_num, 0],
                this_point_fr_xyz[frame_num, 1],
                this_point_fr_xyz[frame_num, 2],
            ]
            bpy.context.view_layer.update()

            this_empty.keyframe_insert(data_path="location", frame=frame_num)

    #######################################################################
    # %% create virtual markers
    left_ear_index = 7
    right_ear_index = 8
    #######################################################################
    print("creating virtual markers")
    print("head_center - midway between left and right ears")
    hips_xyz = (mediapipe_skel_fr_mar_dim[:, left_ear_index, :] + mediapipe_skel_fr_mar_dim[:, right_ear_index, :]) / 2
    bpy.ops.object.empty_add(type="SPHERE")
    this_empty = bpy.context.active_object
    this_empty.name = "head_center"
    this_empty.scale = [empty_size] * 3
    this_empty.parent = freemocap_origin_axes
    mediapipe_tracked_point_names.append(this_empty.name)

    for frame_num in range(end_frame):
        this_empty.location = [
            hips_xyz[frame_num, 0],
            hips_xyz[frame_num, 1],
            hips_xyz[frame_num, 2],
        ]
        this_empty.keyframe_insert(data_path="location", frame=frame_num)

    print("neck_center - midway between left and right shoulders")
    left_shoulder_index = 11
    right_shoulder_index = 12
    neck_xyz = (
        mediapipe_skel_fr_mar_dim[:, left_shoulder_index, :] + mediapipe_skel_fr_mar_dim[:, right_shoulder_index, :]
    ) / 2
    bpy.ops.object.empty_add(type="PLAIN_AXES")
    this_empty = bpy.context.active_object
    this_empty.name = "neck_center"
    this_empty.scale = [empty_size] * 3
    this_empty.parent = freemocap_origin_axes
    mediapipe_tracked_point_names.append(this_empty.name)

    for frame_num in range(end_frame):
        this_empty.location = [
            neck_xyz[frame_num, 0],
            neck_xyz[frame_num, 1],
            neck_xyz[frame_num, 2],
        ]
        this_empty.keyframe_insert(data_path="location", frame=frame_num)

    print("hip_center - midway between left and right hips")
    left_hip_index = 23
    right_hip_index = 24
    hips_xyz = (mediapipe_skel_fr_mar_dim[:, left_hip_index, :] + mediapipe_skel_fr_mar_dim[:, right_hip_index, :]) / 2
    bpy.ops.object.empty_add(type="PLAIN_AXES")
    this_empty = bpy.context.active_object
    this_empty.name = "hip_center"
    this_empty.scale = [empty_size] * 3
    this_empty.parent = freemocap_origin_axes
    mediapipe_tracked_point_names.append(this_empty.name)
    rig_constraint_dict_of_dicts = {
        "spine": {
            "COPY_LOCATION": "hip_center",
            "STRETCH_TO": "chest_center",
        },
        "spine.001": {
            "STRETCH_TO": "neck_center",
        },
        "spine.002": {
            "STRETCH_TO": "head_center",
        },
        "spine.003": {
            "STRETCH_TO": "nose",
        },
        "pelvis.L": {
            "STRETCH_TO": "left_hip",
        },
        "thigh.L": {
            "COPY_LOCATION": "left_hip",
            "STRETCH_TO": "left_knee",
        },
        "shin.L": {
            "STRETCH_TO": "left_ankle",
        },
        "foot.L": {
            "STRETCH_TO": "left_foot_index",
        },
        "heel.02.L": {
            "STRETCH_TO": "left_ankle",
        },
        "shoulder.L": {
            "COPY_LOCATION": "neck_center",
            "STRETCH_TO": "left_shoulder",
        },
        "upper_arm.L": {
            "STRETCH_TO": "left_elbow",
        },
        "forearm.L": {
            "STRETCH_TO": "left_wrist",
        },
        "hand.L": {
            "STRETCH_TO": "left_index",
        },
        "pelvis.R": {
            "STRETCH_TO": "right_hip",
        },
        "thigh.R": {
            "COPY_LOCATION": "right_hip",
            "STRETCH_TO": "right_knee",
        },
        "shin.R": {
            "STRETCH_TO": "right_ankle",
        },
        "foot.R": {
            "STRETCH_TO": "right_foot_index",
        },
        "heel.02.R": {
            "STRETCH_TO": "right_ankle",
        },
        "shoulder.R": {
            "COPY_LOCATION": "neck_center",
            "STRETCH_TO": "right_shoulder",
        },
        "upper_arm.R": {
            "STRETCH_TO": "right_elbow",
        },
        "forearm.R": {
            "STRETCH_TO": "right_wrist",
        },
        "hand.R": {
            "STRETCH_TO": "right_index",
        },
        "thumb.01.R": {
            "COPY_LOCATION": "right_wrist",
            "STRETCH_TO": "right_hand_thumb_mcp",
        },
        "thumb.02.R": {
            "STRETCH_TO": "right_hand_thumb_ip",
        },
        "thumb.02.R": {
            "STRETCH_TO": "right_hand_thumb_tip",
        },
        "palm.01.R": {
            "COPY_LOCATION": "right_wrist",
            "STRETCH_TO": "right_hand_index_finger_mcp",
        },
        "f_index.01.R": {
            "STRETCH_TO": "right_hand_index_finger_pip",
        },
        "f_index.02.R": {
            "STRETCH_TO": "right_hand_index_finger_dip",
        },
        "f_index.03.R": {
            "STRETCH_TO": "right_hand_index_finger_tip",
        },
        "palm.02.R": {
            "COPY_LOCATION": "right_wrist",
            "STRETCH_TO": "right_hand_middle_finger_mcp",
        },
        "f_middle.01.R": {
            "STRETCH_TO": "right_hand_middle_finger_pip",
        },
        "f_middle.02.R": {
            "STRETCH_TO": "right_hand_middle_finger_dip",
        },
        "f_middle.03.R": {
            "STRETCH_TO": "right_hand_middle_finger_tip",
        },
        "palm.03.R": {
            "COPY_LOCATION": "right_wrist",
            "STRETCH_TO": "right_hand_ring_finger_mcp",
        },
        "f_ring.01.R": {
            "STRETCH_TO": "right_hand_ring_finger_pip",
        },
        "f_ring.02.R": {
            "STRETCH_TO": "right_hand_ring_finger_dip",
        },
        "f_ring.03.R": {
            "STRETCH_TO": "right_hand_ring_finger_tip",
        },
        "palm.04.R": {
            "COPY_LOCATION": "right_wrist",
            "STRETCH_TO": "right_hand_pinky_finger_mcp",
        },
        "f_pinky.01.R": {
            "STRETCH_TO": "right_hand_pinky_finger_pip",
        },
        "f_pinky.02.R": {
            "STRETCH_TO": "right_hand_pinky_finger_dip",
        },
        "f_pinky.03.R": {
            "STRETCH_TO": "right_hand_pinky_finger_tip",
        },
        "thumb.01.L": {
            "COPY_LOCATION": "left_wrist",
            "STRETCH_TO": "left_hand_thumb_mcp",
        },
        "thumb.02.L": {
            "STRETCH_TO": "left_hand_thumb_ip",
        },
        "thumb.03.L": {
            "STRETCH_TO": "left_hand_thumb_tip",
        },
        "palm.01.L": {
            "COPY_LOCATION": "left_wrist",
            "STRETCH_TO": "left_hand_index_finger_mcp",
        },
        "f_index.01.L": {
            "STRETCH_TO": "left_hand_index_finger_pip",
        },
        "f_index.02.L": {
            "STRETCH_TO": "left_hand_index_finger_dip",
        },
        "f_index.03.L": {
            "STRETCH_TO": "left_hand_index_finger_tip",
        },
        "palm.02.L": {
            "COPY_LOCATION": "left_wrist",
            "STRETCH_TO": "left_hand_middle_finger_mcp",
        },
        "f_middle.01.L": {
            "STRETCH_TO": "left_hand_middle_finger_pip",
        },
        "f_middle.02.L": {
            "STRETCH_TO": "left_hand_middle_finger_dip",
        },
        "f_middle.03.L": {
            "STRETCH_TO": "left_hand_middle_finger_tip",
        },
        "palm.03.L": {
            "COPY_LOCATION": "left_wrist",
            "STRETCH_TO": "left_hand_ring_finger_mcp",
        },
        "f_ring.01.L": {
            "STRETCH_TO": "left_hand_ring_finger_pip",
        },
        "f_ring.02.L": {
            "STRETCH_TO": "left_hand_ring_finger_dip",
        },
        "f_ring.03.L": {
            "STRETCH_TO": "left_hand_ring_finger_tip",
        },
        "palm.04.L": {
            "COPY_LOCATION": "left_wrist",
            "STRETCH_TO": "left_hand_pinky_finger_mcp",
        },
        "f_pinky.01.L": {
            "STRETCH_TO": "left_hand_pinky_finger_pip",
        },
        "f_pinky.02.L": {
            "STRETCH_TO": "left_hand_pinky_finger_dip",
        },
        "f_pinky.03.L": {
            "STRETCH_TO": "left_hand_pinky_finger_tip",
        },
    }

    # loop through dictionary applying  constraints
    try:
        bpy.ops.object.mode_set(mode="POSE")
    except:
        pass

        for frame_num in range(end_frame):
            this_empty.location = [
                hips_xyz[frame_num, 0],
                hips_xyz[frame_num, 1],
                hips_xyz[frame_num, 2],
            ]
            this_empty.keyframe_insert(data_path="location", frame=frame_num)

        print("chest_center - mean of R/L shoulders and R/L hips")
        chest_xyz = (neck_xyz + hips_xyz) / 2
        bpy.ops.object.empty_add(type="PLAIN_AXES")
        this_empty = bpy.context.active_object
        this_empty.name = "chest_center"
        this_empty.scale = [empty_size] * 3
        this_empty.parent = freemocap_origin_axes
        mediapipe_tracked_point_names.append(this_empty.name)

        for frame_num in range(end_frame):
            this_empty.location = [
                chest_xyz[frame_num, 0],
                chest_xyz[frame_num, 1],
                chest_xyz[frame_num, 2],
            ]
            this_empty.keyframe_insert(data_path="location", frame=frame_num)

        print("Done loading empties :D")

        ######################################################################
        ## Fit Rigify metarig armatures to empties

        bpy.context.scene.frame_set(good_clean_frame_number)

        # each bone  name has to match one in the Blender Rigify Human Metarig, or it will be ignored
        metarig_bone_empty_pairs_dict_of_dicts = {
            "nose.001": {
                "head": "head_center",
                "tail": "nose",
            },
            "nose.L.001": {
                "head": "nose",
                "tail": "left_eye_inner",
            },
            "eye.L": {
                "head": "left_eye_inner",
                "tail": "left_eye",
            },
            "lid.B.L": {
                "head": "left_eye",
                "tail": "left_eye_outer",
            },
            "ear.L": {
                "head": "left_eye_outer",
                "tail": "left_ear",
            },
            "nose.R.001": {
                "head": "nose",
                "tail": "right_eye_inner",
            },
            "eye.R": {
                "head": "right_eye_inner",
                "tail": "right_eye",
            },
            "lid.B.R": {
                "head": "right_eye",
                "tail": "right_eye_outer",
            },
            "ear.R": {
                "head": "right_eye_outer",
                "tail": "right_ear",
            },
            "lip.T.R": {
                "head": "nose",
                "tail": "mouth_right",
            },
            "lip.B.R": {
                "head": "mouth_right",
                "tail": "mouth_left",
            },
            "lip.T.L": {
                "head": "nose",
                "tail": "mouth_left",
            },
            "hand.L": {
                "head": "left_wrist",
                "tail": "left_index",
            },
            "forearm.L": {
                "head": "left_elbow",
                "tail": "left_wrist",
            },
            "upper_arm.L": {
                "head": "left_shoulder",
                "tail": "left_elbow",
            },
            "shoulder.L": {
                "head": "neck_center",
                "tail": "left_shoulder",
            },
            "hand.R": {
                "head": "right_wrist",
                "tail": "right_index",
            },
            "forearm.R": {
                "head": "right_elbow",
                "tail": "right_wrist",
            },
            "upper_arm.R": {
                "head": "right_shoulder",
                "tail": "right_elbow",
            },
            "shoulder.R": {
                "head": "neck_center",
                "tail": "right_shoulder",
            },
            "thigh.L": {
                "head": "left_hip",
                "tail": "left_knee",
            },
            "shin.L": {
                "head": "left_knee",
                "tail": "left_ankle",
            },
            "heel.02.L": {
                "head": "left_heel",
                "tail": "left_ankle",
            },
            "foot.L": {
                "head": "left_ankle",
                "tail": "left_foot_index",
            },
            "toe.L": {
                "head": "left_foot_index",
                "tail": "left_heel",
            },
            "thigh.R": {
                "head": "right_hip",
                "tail": "right_knee",
            },
            "shin.R": {
                "head": "right_knee",
                "tail": "right_ankle",
            },
            "heel.02.R": {
                "head": "right_heel",
                "tail": "right_ankle",
            },
            "foot.R": {
                "head": "right_ankle",
                "tail": "right_foot_index",
            },
            "toe.R": {
                "head": "right_foot_index",
                "tail": "right_heel",
            },
            "pelvis.L": {
                "head": "hip_center",
                "tail": "left_hip",
            },
            "pelvis.R": {
                "head": "hip_center",
                "tail": "right_hip",
            },
            "spine": {
                "head": "hip_center",
                "tail": "chest_center",
            },
            "spine.001": {
                "head": "chest_center",
                "tail": "neck_center",
            },
            "spine.002": {
                "head": "neck_center",
                "tail": "head_center",
            },
            "spine.003": {
                "head": "head_center",
                "tail": "nose",
            },
            "palm.01.R": {
                "head": "right_wrist",
                "tail": "right_hand_index_finger_mcp",
            },
            "palm.02.R": {
                "head": "right_wrist",
                "tail": "right_hand_middle_finger_mcp",
            },
            "palm.03.R": {
                "head": "right_wrist",
                "tail": "right_hand_ring_finger_mcp",
            },
            "palm.04.R": {
                "head": "right_wrist",
                "tail": "right_hand_pinky_finger_mcp",
            },
            "thumb.01.R": {
                "head": "right_wrist",
                "tail": "right_hand_thumb_mcp",
            },
            "thumb.02.R": {
                "head": "right_hand_thumb_mcp",
                "tail": "right_hand_thumb_ip",
            },
            "thumb.03.R": {
                "head": "right_hand_thumb_ip",
                "tail": "right_hand_thumb_tip",
            },
            "f_index.01.R": {
                "head": "right_hand_index_finger_mcp",
                "tail": "right_hand_index_finger_pip",
            },
            "f_index.02.R": {
                "head": "right_hand_index_finger_pip",
                "tail": "right_hand_index_finger_dip",
            },
            "f_index.03.R": {
                "head": "right_hand_index_finger_dip",
                "tail": "right_hand_index_finger_tip",
            },
            "f_middle.01.R": {
                "head": "right_hand_middle_finger_mcp",
                "tail": "right_hand_middle_finger_pip",
            },
            "f_middle.02.R": {
                "head": "right_hand_middle_finger_pip",
                "tail": "right_hand_middle_finger_dip",
            },
            "f_middle.03.R": {
                "head": "right_hand_middle_finger_dip",
                "tail": "right_hand_middle_finger_tip",
            },
            "f_ring.01.R": {
                "head": "right_hand_ring_finger_mcp",
                "tail": "right_hand_ring_finger_pip",
            },
            "f_ring.02.R": {
                "head": "right_hand_ring_finger_pip",
                "tail": "right_hand_ring_finger_dip",
            },
            "f_ring.03.R": {
                "head": "right_hand_ring_finger_dip",
                "tail": "right_hand_ring_finger_tip",
            },
            "f_pinky.01.R": {
                "head": "right_hand_pinky_finger_mcp",
                "tail": "right_hand_pinky_finger_pip",
            },
            "f_pinky.02.R": {
                "head": "right_hand_pinky_finger_pip",
                "tail": "right_hand_pinky_finger_dip",
            },
            "f_pinky.03.R": {
                "head": "right_hand_pinky_finger_dip",
                "tail": "right_hand_pinky_finger_tip",
            },
            "thumb.01.L": {
                "head": "left_wrist",
                "tail": "left_hand_thumb_mcp",
            },
            "thumb.02.L": {
                "head": "left_hand_thumb_mcp",
                "tail": "left_hand_thumb_ip",
            },
            "thumb.03.L": {
                "head": "left_hand_thumb_ip",
                "tail": "left_hand_thumb_tip",
            },
            "palm.01.L": {
                "head": "left_wrist",
                "tail": "left_hand_index_finger_mcp",
            },
            "palm.02.L": {
                "head": "left_wrist",
                "tail": "left_hand_middle_finger_mcp",
            },
            "palm.03.L": {
                "head": "left_wrist",
                "tail": "left_hand_ring_finger_mcp",
            },
            "palm.04.L": {
                "head": "left_wrist",
                "tail": "left_hand_pinky_finger_mcp",
            },
            "f_index.01.L": {
                "head": "left_hand_index_finger_mcp",
                "tail": "left_hand_index_finger_pip",
            },
            "f_index.02.L": {
                "head": "left_hand_index_finger_pip",
                "tail": "left_hand_index_finger_dip",
            },
            "f_index.03.L": {
                "head": "left_hand_index_finger_dip",
                "tail": "left_hand_index_finger_tip",
            },
            "f_middle.01.L": {
                "head": "left_hand_middle_finger_mcp",
                "tail": "left_hand_middle_finger_pip",
            },
            "f_middle.02.L": {
                "head": "left_hand_middle_finger_pip",
                "tail": "left_hand_middle_finger_dip",
            },
            "f_middle.03.L": {
                "head": "left_hand_middle_finger_dip",
                "tail": "left_hand_middle_finger_tip",
            },
            "f_ring.01.L": {
                "head": "left_hand_ring_finger_mcp",
                "tail": "left_hand_ring_finger_pip",
            },
            "f_ring.02.L": {
                "head": "left_hand_ring_finger_pip",
                "tail": "left_hand_ring_finger_dip",
            },
            "f_ring.03.L": {
                "head": "left_hand_ring_finger_dip",
                "tail": "left_hand_ring_finger_tip",
            },
            "f_pinky.01.L": {
                "head": "left_hand_pinky_finger_mcp",
                "tail": "left_hand_pinky_finger_pip",
            },
            "f_pinky.02.L": {
                "head": "left_hand_pinky_finger_pip",
                "tail": "left_hand_pinky_finger_dip",
            },
            "f_pinky.03.L": {
                "head": "left_hand_pinky_finger_dip",
                "tail": "left_hand_pinky_finger_tip",
            },
        }

        # %% create metarig
        bpy.ops.object.armature_human_metarig_add()
        this_metarig = bpy.context.active_object

        # %% loop through dictionary - set head and tail of each bone's to the relevant empty
        # bpy.ops.object.posemode_toggle()
        try:
            bpy.ops.object.mode_set(mode="EDIT")
        except:
            pass

        for this_bone in list(this_metarig.data.bones):
            this_bone_name = this_bone.name
            print(f"--{this_bone_name}--")

            this_edit_bone = this_metarig.data.edit_bones[
                this_bone_name
            ]  # note this is using EDIT_bones, not regular bones b/c blender is a wild wacky place

            if this_bone_name in metarig_bone_empty_pairs_dict_of_dicts:
                this_bone_dict = metarig_bone_empty_pairs_dict_of_dicts[this_bone_name]

                this_head_empty_name = this_bone_dict["head"]
                this_tail_empty_name = this_bone_dict["tail"]
                print(
                    "setting (bone:head->tail){}:{}->{})".format(
                        this_bone_name, this_head_empty_name, this_tail_empty_name
                    )
                )

                if this_tail_empty_name is not None:
                    print(
                        f"setting bone: {this_bone_name}: head: {bpy.data.objects[this_head_empty_name].location}, tail: {bpy.data.objects[this_tail_empty_name].location}"
                    )
                    this_edit_bone.head = bpy.data.objects[this_head_empty_name].location
                    this_edit_bone.tail = bpy.data.objects[this_tail_empty_name].location
                else:  # remove bones with no tail (and/or head) until we figure out a better plan
                    print(f"removing bone: {this_bone_name}")
                    this_metarig.data.edit_bones.remove(this_edit_bone)
            else:
                print(f"removing bone: {this_bone_name}")
                this_metarig.data.edit_bones.remove(this_edit_bone)

        try:
            bpy.ops.object.mode_set(mode="OBJECT")
        except:
            pass

        #######################################################
        #### Damped track bones to empties
        bpy.ops.object.mode_set(mode="OBJECT")
        armature_name = "metarig"
        this_armature = bpy.data.objects[armature_name]

        rig_constraint_dict_of_dicts = {
            "nose.001": {
                "COPY_LOCATION": "head_center",
                "DAMPED_TRACK": "nose",
            },
            "nose.L.001": {
                "DAMPED_TRACK": "left_eye_inner",
            },
            "eye.L": {
                "DAMPED_TRACK": "left_eye",
            },
            "lid.B.L": {
                "DAMPED_TRACK": "left_eye_outer",
            },
            "ear.L": {
                "DAMPED_TRACK": "left_ear",
            },
            "nose.R.001": {
                "DAMPED_TRACK": "right_eye_inner",
            },
            "eye.R": {
                "DAMPED_TRACK": "right_eye",
            },
            "lid.B.R": {
                "DAMPED_TRACK": "right_eye_outer",
            },
            "ear.R": {
                "DAMPED_TRACK": "right_ear",
            },
            "lip.T.R": {
                "DAMPED_TRACK": "mouth_right",
            },
            "lip.B.R": {
                "DAMPED_TRACK": "mouth_left",
            },
            "lip.T.L": {
                "DAMPED_TRACK": "nose",
            },
            "spine": {
                "COPY_LOCATION": "hip_center",
                "DAMPED_TRACK": "chest_center",
            },
            "spine.001": {
                "DAMPED_TRACK": "neck_center",
            },
            "spine.002": {
                "DAMPED_TRACK": "head_center",
            },
            "spine.003": {
                "DAMPED_TRACK": "nose",
            },
            "pelvis.L": {
                "DAMPED_TRACK": "left_hip",
            },
            "thigh.L": {
                "COPY_LOCATION": "left_hip",
                "DAMPED_TRACK": "left_knee",
            },
            "shin.L": {
                "DAMPED_TRACK": "left_ankle",
            },
            "foot.L": {
                "DAMPED_TRACK": "left_foot_index",
            },
            "heel.02.L": {
                "DAMPED_TRACK": "left_ankle",
            },
            "shoulder.L": {
                "COPY_LOCATION": "neck_center",
                "DAMPED_TRACK": "left_shoulder",
            },
            "upper_arm.L": {
                "DAMPED_TRACK": "left_elbow",
            },
            "forearm.L": {
                "DAMPED_TRACK": "left_wrist",
            },
            "hand.L": {
                "DAMPED_TRACK": "left_index",
            },
            "pelvis.R": {
                "DAMPED_TRACK": "right_hip",
            },
            "thigh.R": {
                "COPY_LOCATION": "right_hip",
                "DAMPED_TRACK": "right_knee",
            },
            "shin.R": {
                "DAMPED_TRACK": "right_ankle",
            },
            "foot.R": {
                "DAMPED_TRACK": "right_foot_index",
            },
            "heel.02.R": {
                "DAMPED_TRACK": "right_ankle",
            },
            "shoulder.R": {
                "COPY_LOCATION": "neck_center",
                "DAMPED_TRACK": "right_shoulder",
            },
            "upper_arm.R": {
                "DAMPED_TRACK": "right_elbow",
            },
            "forearm.R": {
                "DAMPED_TRACK": "right_wrist",
            },
            "hand.R": {
                "DAMPED_TRACK": "right_index",
            },
            "thumb.01.R": {
                "COPY_LOCATION": "right_wrist",
                "DAMPED_TRACK": "right_hand_thumb_mcp",
            },
            "thumb.02.R": {
                "DAMPED_TRACK": "right_hand_thumb_ip",
            },
            "thumb.02.R": {
                "DAMPED_TRACK": "right_hand_thumb_tip",
            },
            "palm.01.R": {
                "COPY_LOCATION": "right_wrist",
                "DAMPED_TRACK": "right_hand_index_finger_mcp",
            },
            "f_index.01.R": {
                "DAMPED_TRACK": "right_hand_index_finger_pip",
            },
            "f_index.02.R": {
                "DAMPED_TRACK": "right_hand_index_finger_dip",
            },
            "f_index.03.R": {
                "DAMPED_TRACK": "right_hand_index_finger_tip",
            },
            "palm.02.R": {
                "COPY_LOCATION": "right_wrist",
                "DAMPED_TRACK": "right_hand_middle_finger_mcp",
            },
            "f_middle.01.R": {
                "DAMPED_TRACK": "right_hand_middle_finger_pip",
            },
            "f_middle.02.R": {
                "DAMPED_TRACK": "right_hand_middle_finger_dip",
            },
            "f_middle.03.R": {
                "DAMPED_TRACK": "right_hand_middle_finger_tip",
            },
            "palm.03.R": {
                "COPY_LOCATION": "right_wrist",
                "DAMPED_TRACK": "right_hand_ring_finger_mcp",
            },
            "f_ring.01.R": {
                "DAMPED_TRACK": "right_hand_ring_finger_pip",
            },
            "f_ring.02.R": {
                "DAMPED_TRACK": "right_hand_ring_finger_dip",
            },
            "f_ring.03.R": {
                "DAMPED_TRACK": "right_hand_ring_finger_tip",
            },
            "palm.04.R": {
                "COPY_LOCATION": "right_wrist",
                "DAMPED_TRACK": "right_hand_pinky_finger_mcp",
            },
            "f_pinky.01.R": {
                "DAMPED_TRACK": "right_hand_pinky_finger_pip",
            },
            "f_pinky.02.R": {
                "DAMPED_TRACK": "right_hand_pinky_finger_dip",
            },
            "f_pinky.03.R": {
                "DAMPED_TRACK": "right_hand_pinky_finger_tip",
            },
            "thumb.01.L": {
                "COPY_LOCATION": "left_wrist",
                "DAMPED_TRACK": "left_hand_thumb_mcp",
            },
            "thumb.02.L": {
                "DAMPED_TRACK": "left_hand_thumb_ip",
            },
            "thumb.03.L": {
                "DAMPED_TRACK": "left_hand_thumb_tip",
            },
            "palm.01.L": {
                "COPY_LOCATION": "left_wrist",
                "DAMPED_TRACK": "left_hand_index_finger_mcp",
            },
            "f_index.01.L": {
                "DAMPED_TRACK": "left_hand_index_finger_pip",
            },
            "f_index.02.L": {
                "DAMPED_TRACK": "left_hand_index_finger_dip",
            },
            "f_index.03.L": {
                "DAMPED_TRACK": "left_hand_index_finger_tip",
            },
            "palm.02.L": {
                "COPY_LOCATION": "left_wrist",
                "DAMPED_TRACK": "left_hand_middle_finger_mcp",
            },
            "f_middle.01.L": {
                "DAMPED_TRACK": "left_hand_middle_finger_pip",
            },
            "f_middle.02.L": {
                "DAMPED_TRACK": "left_hand_middle_finger_dip",
            },
            "f_middle.03.L": {
                "DAMPED_TRACK": "left_hand_middle_finger_tip",
            },
            "palm.03.L": {
                "COPY_LOCATION": "left_wrist",
                "DAMPED_TRACK": "left_hand_ring_finger_mcp",
            },
            "f_ring.01.L": {
                "DAMPED_TRACK": "left_hand_ring_finger_pip",
            },
            "f_ring.02.L": {
                "DAMPED_TRACK": "left_hand_ring_finger_dip",
            },
            "f_ring.03.L": {
                "DAMPED_TRACK": "left_hand_ring_finger_tip",
            },
            "palm.04.L": {
                "COPY_LOCATION": "left_wrist",
                "DAMPED_TRACK": "left_hand_pinky_finger_mcp",
            },
            "f_pinky.01.L": {
                "DAMPED_TRACK": "left_hand_pinky_finger_pip",
            },
            "f_pinky.02.L": {
                "DAMPED_TRACK": "left_hand_pinky_finger_dip",
            },
            "f_pinky.03.L": {
                "DAMPED_TRACK": "left_hand_pinky_finger_tip",
            },
        }

        # loop through dictionary applying  constraints
        try:
            bpy.ops.object.mode_set(mode="POSE")
        except:
            pass

        for this_bone_name, this_bone_dict in rig_constraint_dict_of_dicts.items():
            print(f"---Setting constraints for bone:{this_bone_name}---")

            for (
                this_constraint_name,
                this_constraint_target_empty_name,
            ) in this_bone_dict.items():
                print(f"constraint: {this_constraint_name} with target:{this_constraint_target_empty_name}")
                print("grab bone")
                this_bone = this_armature.pose.bones[this_bone_name]
                print("apply bone")
                this_constraint = this_bone.constraints.new(type=this_constraint_name)
                this_constraint.name = this_constraint_name
                this_constraint.target = bpy.data.objects[
                    this_constraint_target_empty_name
                ]  # point constraint at relevant empty object

        ##############
        ## create extra bones

        extra_bone_empty_pairs_dict_of_dicts = {
            "hand.L": {
                "head": "left_wrist",
                "tail": "left_index",
            },
            "forearm.L": {
                "head": "left_elbow",
                "tail": "left_wrist",
            },
        }

        ###################################################################
        #### Create Stick Figure Mesh

        stick_figure_mesh_edge_pairs_dict_of_dicts = {
            "nose.001": {
                "head": "head_center",
                "tail": "nose",
            },
            "nose.L.001": {
                "head": "nose",
                "tail": "left_eye_inner",
            },
            "eye.L": {
                "head": "left_eye_inner",
                "tail": "left_eye",
            },
            "lid.B.L": {
                "head": "left_eye",
                "tail": "left_eye_outer",
            },
            "ear.L": {
                "head": "left_eye_outer",
                "tail": "left_ear",
            },
            "nose.R.001": {
                "head": "nose",
                "tail": "right_eye_inner",
            },
            "eye.R": {
                "head": "right_eye_inner",
                "tail": "right_eye",
            },
            "lid.B.R": {
                "head": "right_eye",
                "tail": "right_eye_outer",
            },
            "ear.R": {
                "head": "right_eye_outer",
                "tail": "right_ear",
            },
            "lip.T.R": {
                "head": "nose",
                "tail": "mouth_right",
            },
            "lip.B.R": {
                "head": "mouth_right",
                "tail": "mouth_left",
            },
            "lip.T.L": {
                "head": "nose",
                "tail": "mouth_left",
            },
            "forearm.L": {
                "head": "left_elbow",
                "tail": "left_wrist",
            },
            "upper_arm.L": {
                "head": "left_shoulder",
                "tail": "left_elbow",
            },
            "shoulder.L": {
                "head": "neck_center",
                "tail": "left_shoulder",
            },
            "forearm.R": {
                "head": "right_elbow",
                "tail": "right_wrist",
            },
            "upper_arm.R": {
                "head": "right_shoulder",
                "tail": "right_elbow",
            },
            "shoulder.R": {
                "head": "neck_center",
                "tail": "right_shoulder",
            },
            "thigh.L": {
                "head": "left_hip",
                "tail": "left_knee",
            },
            "shin.L": {
                "head": "left_knee",
                "tail": "left_ankle",
            },
            "heel.02.L": {
                "head": "left_heel",
                "tail": "left_ankle",
            },
            "foot.L": {
                "head": "left_ankle",
                "tail": "left_foot_index",
            },
            "toe.L": {
                "head": "left_foot_index",
                "tail": "left_heel",
            },
            "thigh.R": {
                "head": "right_hip",
                "tail": "right_knee",
            },
            "shin.R": {
                "head": "right_knee",
                "tail": "right_ankle",
            },
            "heel.02.R": {
                "head": "right_heel",
                "tail": "right_ankle",
            },
            "foot.R": {
                "head": "right_ankle",
                "tail": "right_foot_index",
            },
            "toe.R": {
                "head": "right_foot_index",
                "tail": "right_heel",
            },
            "pelvis.L": {
                "head": "hip_center",
                "tail": "left_hip",
            },
            "pelvis.R": {
                "head": "hip_center",
                "tail": "right_hip",
            },
            "spine": {
                "head": "hip_center",
                "tail": "chest_center",
            },
            "spine.001": {
                "head": "chest_center",
                "tail": "neck_center",
            },
            "spine.002": {
                "head": "neck_center",
                "tail": "head_center",
            },
            "spine.003": {
                "head": "head_center",
                "tail": "nose",
            },
            "palm.01.R": {
                "head": "right_wrist",
                "tail": "right_hand_index_finger_mcp",
            },
            "palm.02.R": {
                "head": "right_wrist",
                "tail": "right_hand_middle_finger_mcp",
            },
            "palm.03.R": {
                "head": "right_wrist",
                "tail": "right_hand_ring_finger_mcp",
            },
            "palm.04.R": {
                "head": "right_wrist",
                "tail": "right_hand_pinky_finger_mcp",
            },
            "thumb.01.R": {
                "head": "right_wrist",
                "tail": "right_hand_thumb_cmc",
            },
            "thumb.02.R": {
                "head": "right_hand_thumb_cmc",
                "tail": "right_hand_thumb_mcp",
            },
            "thumb.03.R": {
                "head": "right_hand_thumb_mcp",
                "tail": "right_hand_thumb_ip",
            },
            "thumb.04.R": {
                "head": "right_hand_thumb_ip",
                "tail": "right_hand_thumb_tip",
            },
            "f_index.01.R": {
                "head": "right_hand_index_finger_mcp",
                "tail": "right_hand_index_finger_pip",
            },
            "f_index.02.R": {
                "head": "right_hand_index_finger_pip",
                "tail": "right_hand_index_finger_dip",
            },
            "f_index.03.R": {
                "head": "right_hand_index_finger_dip",
                "tail": "right_hand_index_finger_tip",
            },
            "f_middle.01.R": {
                "head": "right_hand_middle_finger_mcp",
                "tail": "right_hand_middle_finger_pip",
            },
            "f_middle.02.R": {
                "head": "right_hand_middle_finger_pip",
                "tail": "right_hand_middle_finger_dip",
            },
            "f_middle.03.R": {
                "head": "right_hand_middle_finger_dip",
                "tail": "right_hand_middle_finger_tip",
            },
            "f_ring.01.R": {
                "head": "right_hand_ring_finger_mcp",
                "tail": "right_hand_ring_finger_pip",
            },
            "f_ring.02.R": {
                "head": "right_hand_ring_finger_pip",
                "tail": "right_hand_ring_finger_dip",
            },
            "f_ring.03.R": {
                "head": "right_hand_ring_finger_dip",
                "tail": "right_hand_ring_finger_tip",
            },
            "f_pinky.01.R": {
                "head": "right_hand_pinky_finger_mcp",
                "tail": "right_hand_pinky_finger_pip",
            },
            "f_pinky.02.R": {
                "head": "right_hand_pinky_finger_pip",
                "tail": "right_hand_pinky_finger_dip",
            },
            "f_pinky.03.R": {
                "head": "right_hand_pinky_finger_dip",
                "tail": "right_hand_pinky_finger_tip",
            },
            "thumb.01.L": {
                "head": "left_wrist",
                "tail": "left_hand_thumb_cmc",
            },
            "thumb.02.L": {
                "head": "left_hand_thumb_cmc",
                "tail": "left_hand_thumb_mcp",
            },
            "thumb.03.L": {
                "head": "left_hand_thumb_mcp",
                "tail": "left_hand_thumb_ip",
            },
            "thumb.04.L": {
                "head": "left_hand_thumb_ip",
                "tail": "left_hand_thumb_tip",
            },
            "palm.01.L": {
                "head": "left_wrist",
                "tail": "left_hand_index_finger_mcp",
            },
            "palm.02.L": {
                "head": "left_wrist",
                "tail": "left_hand_middle_finger_mcp",
            },
            "palm.03.L": {
                "head": "left_wrist",
                "tail": "left_hand_ring_finger_mcp",
            },
            "palm.04.L": {
                "head": "left_wrist",
                "tail": "left_hand_pinky_finger_mcp",
            },
            "f_index.01.L": {
                "head": "left_hand_index_finger_mcp",
                "tail": "left_hand_index_finger_pip",
            },
            "f_index.02.L": {
                "head": "left_hand_index_finger_pip",
                "tail": "left_hand_index_finger_dip",
            },
            "f_index.03.L": {
                "head": "left_hand_index_finger_dip",
                "tail": "left_hand_index_finger_tip",
            },
            "f_middle.01.L": {
                "head": "left_hand_middle_finger_mcp",
                "tail": "left_hand_middle_finger_pip",
            },
            "f_middle.02.L": {
                "head": "left_hand_middle_finger_pip",
                "tail": "left_hand_middle_finger_dip",
            },
            "f_middle.03.L": {
                "head": "left_hand_middle_finger_dip",
                "tail": "left_hand_middle_finger_tip",
            },
            "f_ring.01.L": {
                "head": "left_hand_ring_finger_mcp",
                "tail": "left_hand_ring_finger_pip",
            },
            "f_ring.02.L": {
                "head": "left_hand_ring_finger_pip",
                "tail": "left_hand_ring_finger_dip",
            },
            "f_ring.03.L": {
                "head": "left_hand_ring_finger_dip",
                "tail": "left_hand_ring_finger_tip",
            },
            "f_pinky.01.L": {
                "head": "left_hand_pinky_finger_mcp",
                "tail": "left_hand_pinky_finger_pip",
            },
            "f_pinky.02.L": {
                "head": "left_hand_pinky_finger_pip",
                "tail": "left_hand_pinky_finger_dip",
            },
            "f_pinky.03.L": {
                "head": "left_hand_pinky_finger_dip",
                "tail": "left_hand_pinky_finger_tip",
            },
        }

        mediapipe_virtual_markers_weights_dict = {
            "head_center": {"right_ear": 0.5, "left_ear": 0.5},
            "neck_center": {"right_shoulder": 0.5, "left_shoulder": 0.5},
            "hip_center": {"right_hip": 0.5, "left_hip": 0.5},
            "chest_center": {
                "right_shoulder": 0.25,
                "left_shoulder": 0.25,
                "right_hip": 0.25,
                "left_hip": 0.25,
            },
        }

        print("loading data as stick figure mesh \o/")
        #######################################################################

        # %% create virtual marker points (redundant from above, I think ? )

        current_frame = bpy.data.scenes[0].frame_current
        virtual_point_index_dict = {}
        for this_virtual_marker_dict in mediapipe_virtual_markers_weights_dict.items():
            this_virtual_marker_name = this_virtual_marker_dict[0]
            these_weighted_points = []
            for this_weighted_point_dict in this_virtual_marker_dict[1].items():
                this_point_name = this_weighted_point_dict[0]

                this_weight = this_weighted_point_dict[1]
                these_weighted_points.append(
                    mediapipe_skel_fr_mar_dim[:, mediapipe_tracked_point_names.index(this_point_name), :] * this_weight
                )
            this_virtual_point_fr_xyz = np.sum(these_weighted_points, axis=0)
            this_virtual_point_fr_mar_xyz = np.expand_dims(this_virtual_point_fr_xyz, axis=1)
            mediapipe_skel_fr_mar_dim = np.append(mediapipe_skel_fr_mar_dim, this_virtual_point_fr_mar_xyz, axis=1)
            virtual_point_index_dict[this_virtual_marker_name] = mediapipe_skel_fr_mar_dim.shape[1] - 1
            print(
                f"Virtual marker: {this_virtual_marker_name} created at marker index {mediapipe_skel_fr_mar_dim.shape[1] - 1}"
            )

        # %% define edges

        edges = []
        # regular data
        for this_bone_dict in stick_figure_mesh_edge_pairs_dict_of_dicts.items():
            edge_name = this_bone_dict[0]
            head_name = this_bone_dict[1]["head"]
            tail_name = this_bone_dict[1]["tail"]

            try:

                head_index = mediapipe_tracked_point_names.index(head_name)
                tail_index = mediapipe_tracked_point_names.index(tail_name)

                if (
                    "center" in head_name
                ):  # virtual markers are acting weird so we're brute forcing it :-/ this will fail if there is a virtual marker without 'center' in the name....
                    head_index = virtual_point_index_dict[head_name]
                if "center" in tail_name:
                    tail_index = virtual_point_index_dict[tail_name]

                this_edge = (head_index, tail_index)

                edges.append(this_edge)
                print(
                    f"edge created for bone: ({this_bone_dict[0]} : {head_name}-{tail_name}), indicies({this_edge[0]},{this_edge[1]})"
                )
            except:
                print(f"edge FAILED for for bone:{this_bone_dict[0]} : {head_name}-{tail_name}")
                pass

        print("edges created!")
        # #
        # # ##############################
        # # # %% Create mesh
        # # try:
        # #     bpy.ops.object.mode_set(mode='OBJECT')
        # # except:
        # #     pass
        # #
        # # print('loading verticies')
        # # vertices = mediapipe_skel_fr_mar_dim[current_frame, :,
        # #            :]  # don't plot face dottos until we know what to do with them
        # # print('done loading verticies')
        # # # edges defined above
        # # faces = []
        # # stick_figure_mesh = bpy.data.meshes.new('stick_figure_mesh')
        # # print('a')
        # # stick_figure_mesh.from_pydata(vertices, edges, faces)
        # # print('b')
        # # stick_figure_mesh.update()
        # # print('c')
        # # # make object from mesh
        # # stick_figure_mesh_object = bpy.data.objects.new('stick_figure_mesh', stick_figure_mesh)
        # # print('d')
        # # # make collection
        # # mesh_collection = bpy.data.collections.new('mesh_collection')
        # # print('e')
        # # bpy.context.scene.collection.children.link(mesh_collection)
        # # print('f')
        # # # add object to scene collection
        # # mesh_collection.objects.link(stick_figure_mesh_object)
        # # print('g')
        # #
        # # bpy.context.view_layer.objects.active = bpy.data.objects['stick_figure_mesh']
        # # print('h')
        #
        # bpy.context.object.mode
        # try:
        #     bpy.ops.object.editmode_toggle()
        # except:
        #     pass
        # print('i')
        #
        # #### skin that mesh!
        # print('skinning mesh \o/')
        # bpy.data.objects['stick_figure_mesh'].modifiers.new('stick_figure_mesh_skin', 'SKIN')
        # print('resizing mesh')
        # skin_radius = .025
        # bpy.ops.transform.skin_resize(value=(skin_radius, skin_radius, skin_radius))
        #
        # ########
        # # # parent mesh to armature
        # print('parenting mesh to armarture with automatic weights...')
        # bpy.ops.object.mode_set(mode='OBJECT')
        # bpy.data.objects['stick_figure_mesh'].parent = this_metarig
        # bpy.context.view_layer.objects.active = this_metarig
        # bpy.context.active_object.select_set(True)
        # bpy.data.objects['stick_figure_mesh'].select_set(True)
        # bpy.ops.object.parent_set(type='ARMATURE_AUTO')

        #####################
        ## Load nSynched Videos
        try:
            print("loading videos as planes...")
            annotated_videos_path = session_path / "annotated_videos"

            if annotated_videos_path.is_dir():
                vidFolderPath = annotated_videos_path
            else:
                vidFolderPath = session_path / "synchronized_videos"

            world_origin = bpy.data.objects["world_origin"]

            number_of_videos = len(list(get_video_paths(vidFolderPath)))

            vid_location_scale = 1

            for (
                vid_number,
                thisVidPath,
            ) in enumerate(get_video_paths(vidFolderPath)):
                print(thisVidPath)
                # use 'images as planes' add on to load in the video files as planes
                bpy.ops.import_image.to_plane(
                    files=[{"name": thisVidPath.name}],
                    directory=str(thisVidPath.parent),
                    shader="EMISSION",
                )
                this_vid_as_plane = bpy.context.active_object
                this_vid_as_plane.name = "vid_" + str(vid_number)

                vid_x = (vid_number - number_of_videos / 2) * vid_location_scale

                this_vid_as_plane.location = [
                    vid_x,
                    vid_location_scale,
                    vid_location_scale,
                ]
                this_vid_as_plane.rotation_euler = [np.pi / 2, 0, 0]
                this_vid_as_plane.scale = [vid_location_scale * 1.5] * 3
                this_vid_as_plane.parent = world_origin
                # create a light
                # bpy.ops.object.light_add(type='POINT', radius=1, align='WORLD')
        except:
            print(
                'Failed to load videos to Blender scene - You might need to install the "images as planes" addon to this version of Blender'
            )
except Exception as e:
    print(e)
    print(
        "Something broke in the Blender Megascript - Maybe it failed to find a good clean frame and then that borked something downstream?"
    )

# set shading type to 'Material Preview' so you can see the videos
# bpy.context.space_data.shading.type = "MATERIAL"

# save .blend file
sessionID = session_path.stem

print(f"Saving Blender output file to: {str(blend_file_save_path)}")
bpy.ops.wm.save_as_mainfile(filepath=str(blend_file_save_path))



"""Top-level package for freemocap"""

__author__ = """Skelly FreeMoCap"""
__email__ = "info@freemocap.org"
__version__ = "v1.0.7-rc"
__description__ = "A free and open source markerless motion capture system for everyone 💀✨"

__package_name__ = "freemocap"
__repo_url__ = f"https://github.com/freemocap/{__package_name__}/"
__repo_issues_url__ = f"{__repo_url__}issues"

import sys
from pathlib import Path

base_package_path = Path(__file__).parent.parent
print(f"adding base_package_path: {base_package_path} : to sys.path")
sys.path.insert(0, str(base_package_path))  # add par

from freemocap.system.logging.configure_logging import configure_logging

configure_logging()
import logging

logger = logging.getLogger(__name__)
logger.info(f"Initializing {__package_name__} package, version: {__version__}, from file: {__file__}")


def RunMe(*args, **kwargs):
    logger.info(
        "User tried using `pre-alpha` entry point (`import freemocap: freemocap.RunMe() - displaying friendly message then re-directing to `freemocap.__main__:main()` entry point")

    print("--------------------------------\n"
          "--------------------------------\n"
          "--------------------------------\n"
          "Hello! Looks like you're trying to use the `pre-alpha` entry point for FreeMoCap.\n"
          "This entry point is deprecated, so we're launching the GUI via `freemocap.__main__:main()` entry point.\n"
          "if you need to install the `pre-alpha` code, use `pip install freemocap==0.0.54`"
          "Thank you for using FreeMoCap!\n"
          "--------------------------------\n"
          "--------------------------------\n"
          "(NOTE  - this entry point will be removed eventually\n"
          "--------------------------------\n")

    from freemocap.__main__ import main
    main()



from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QGroupBox,
    QVBoxLayout,
    QHBoxLayout,
    QLineEdit,
    QFormLayout,
    QLabel,
    QRadioButton,
    QCheckBox,
)
from skellycam import SkellyCamControllerWidget

from freemocap.core_processes.capture_volume_calibration.charuco_stuff.default_charuco_square_size import (
    default_charuco_square_size_mm,
)
from freemocap.system.paths_and_files_names import create_new_default_recording_name, create_new_recording_folder_path


class RecordingNameGenerator:
    pass


class CameraControllerGroupBox(QGroupBox):
    def __init__(self, skellycam_controller: SkellyCamControllerWidget, parent=None):
        super().__init__(parent=parent)
        self._skellycam_controller = skellycam_controller
        skellycam_controller.start_recording_button.setObjectName("start_recording_button")
        skellycam_controller.stop_recording_button.setObjectName("stop_recording_button")

        # self.setFlat(True)
        self._layout = QVBoxLayout()
        self._layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setLayout(self._layout)

        # self._layout.setContentsMargins(0, 0, 0, 0)
        # self._layout.setSpacing(0)
        self._layout.setAlignment(Qt.AlignmentFlag.AlignTop)



        motion_capture_recording_options_layout = self._create_mocap_recording_option_layout()
        self._layout.addLayout(motion_capture_recording_options_layout)

        calibration_recording_option_layout = self._create_calibration_recording_option_layout()
        self._layout.addLayout(calibration_recording_option_layout)

        self._layout.addLayout(self._create_videos_will_save_to_layout())

        self._layout.addWidget(self._skellycam_controller)

    def _create_videos_will_save_to_layout(self):
        vbox = QVBoxLayout()

        hbox = QHBoxLayout()
        self._recording_string_tag_line_edit = QLineEdit(parent=self)
        self._recording_string_tag_line_edit.setPlaceholderText("(Optional)")

        recording_string_tag_form_layout = QFormLayout(parent=self)
        recording_string_tag_form_layout.addRow("Tag:", self._recording_string_tag_line_edit)
        hbox.addLayout(recording_string_tag_form_layout)

        hbox.setAlignment(Qt.AlignmentFlag.AlignLeft)
        videos_will_save_to_label = QLabel("Videos will save to folder: ")
        videos_will_save_to_label.setStyleSheet("font-size: 10px;")
        hbox.addWidget(videos_will_save_to_label)
        self._recording_path_label = QLabel(f"{self.get_new_recording_path()}")
        self._recording_path_label.setStyleSheet("font-family: monospace; font-size: 10px;")
        hbox.addWidget(self._recording_path_label)

        vbox.addLayout(hbox)

        lag_note_label = QLabel("NOTE: If you experience lag in your camera views, decrease the resolution and/or use fewer cameras.\nWe are working on a fix which should be done 'soon' (written: 2023-04-05)")
        vbox.addWidget(lag_note_label)

        return vbox

    def _create_mocap_recording_option_layout(self):
        hbox = QHBoxLayout()
        hbox.setAlignment(Qt.AlignmentFlag.AlignLeft)

        self._mocap_videos_radio_button = QRadioButton("Record Motion Capture Videos")
        hbox.addWidget(self._mocap_videos_radio_button)

        self._mocap_videos_radio_button.setChecked(True)
        hbox.addWidget(QLabel(" - "))
        self._auto_process_videos_checkbox = QCheckBox("Auto Process Videos on Save")
        self._auto_process_videos_checkbox.setChecked(True)
        hbox.addWidget(self._auto_process_videos_checkbox)
        self._auto_open_in_blender_checkbox = QCheckBox("Auto Open in Blender")
        self._auto_open_in_blender_checkbox.setChecked(True)
        hbox.addWidget(self._auto_open_in_blender_checkbox)
        return hbox

    def _create_calibration_recording_option_layout(self):
        hbox = QHBoxLayout()
        hbox.setAlignment(Qt.AlignmentFlag.AlignLeft)

        self._calibration_videos_radio_button = QRadioButton("Record Calibration Videos")
        hbox.addWidget(self._calibration_videos_radio_button)
        hbox.addWidget(QLabel(" - "))
        self._calibration_videos_radio_button.toggled.connect(self._handle_calibration_videos_radio_button_changed)

        hbox.addWidget(QLabel("Charuco square size (mm)", parent=self))
        self._charuco_square_size_line_edit = QLineEdit(parent=self)
        self._charuco_square_size_line_edit.setFixedWidth(100)
        self._charuco_square_size_line_edit.setText(str(default_charuco_square_size_mm))
        self._charuco_square_size_line_edit.setToolTip(
            "The length of one of the edges of the black squares in the calibration board in mm"
        )
        hbox.addWidget(self._charuco_square_size_line_edit)
        hbox.addStretch()
        return hbox

    @property
    def mocap_videos_radio_button_checked(self) -> bool:
        return self._mocap_videos_radio_button.isChecked()

    @property
    def calibration_videos_radio_button_checked(self) -> bool:
        return self._calibration_videos_radio_button.isChecked()

    @property
    def auto_process_videos_checked(self) -> bool:
        return self._auto_process_videos_checkbox.isChecked()

    @property
    def auto_open_in_blender_checked(self) -> bool:
        return self._auto_open_in_blender_checkbox.isChecked()

    @property
    def charuco_square_size(self) -> float:
        return float(self._charuco_square_size_line_edit.text())

    def get_new_recording_path(self):
        return create_new_recording_folder_path(recording_name=self._get_recording_name())

    def update_recording_name_string(self):
        self._recording_path_label.setText(self.get_new_recording_path())

    def _get_recording_name_string_tag(self):
        try:
            tag = self._recording_string_tag_line_edit.text()
            tag = tag.replace("   ", " ")
            tag = tag.replace("  ", " ")
            tag = tag.replace(" ", "_")
            return tag
        except:
            return ""

    def _get_recording_name(self):
        name = create_new_default_recording_name()
        if self._calibration_videos_radio_button.isChecked():
            name = f"{name}_calibration"

        tag = self._get_recording_name_string_tag()
        if tag == "":
            return name
        else:
            return f"{name}__{self._get_recording_name_string_tag()}"

    def _handle_calibration_videos_radio_button_changed(self, state):
        if self._calibration_videos_radio_button.isChecked():
            self.setProperty("calibration_videos_radio_button_checked", True)
            self._skellycam_controller.set_calibration_recordings_button_label(True)
        else:
            self.setProperty("calibration_videos_radio_button_checked", False)
            self._skellycam_controller.set_calibration_recordings_button_label(False)
        self.style().polish(self)



import logging
from dataclasses import dataclass
import multiprocessing
from pathlib import Path
from typing import Any, List, Union

import cv2
import mediapipe as mp
import numpy as np
from skellycam.detection.models.frame_payload import FramePayload
from skellycam.opencv.video_recorder.video_recorder import VideoRecorder
from tqdm import tqdm

from freemocap.system.paths_and_files_names import MEDIAPIPE_2D_NPY_FILE_NAME, ANNOTATED_VIDEOS_FOLDER_NAME, \
    MEDIAPIPE_BODY_WORLD_FILE_NAME
from freemocap.core_processes.detecting_things_in_2d_images.mediapipe_stuff.mediapipe_skeleton_names_and_connections import (
    mediapipe_tracked_point_names_dict,
)
from freemocap.parameter_info_models.recording_processing_parameter_models import MediapipeParametersModel
from freemocap.utilities.get_video_paths import get_video_paths


logger = logging.getLogger(__name__)


@dataclass
class Mediapipe2dNumpyArrays:
    body_frameNumber_trackedPointNumber_XYZ: np.ndarray = None
    body_world_frameNumber_trackedPointNumber_XYZ: np.ndarray = None
    rightHand_frameNumber_trackedPointNumber_XYZ: np.ndarray = None
    leftHand_frameNumber_trackedPointNumber_XYZ: np.ndarray = None
    face_frameNumber_trackedPointNumber_XYZ: np.ndarray = None

    body_frameNumber_trackedPointNumber_confidence: np.ndarray = None

    @property
    def has_data(self):
        return not np.isnan(self.body_frameNumber_trackedPointNumber_XYZ).all()

    @property
    def all_data2d_nFrames_nTrackedPts_XY(self):
        """dimensions will be [number_of_frames , number_of_markers, XY]"""

        if self.body_frameNumber_trackedPointNumber_XYZ is None:
            # if there's no body data, there's no hand or face data either
            return

        if len(self.body_frameNumber_trackedPointNumber_XYZ.shape) == 3:  # multiple frames
            return np.hstack(
                [
                    self.body_frameNumber_trackedPointNumber_XYZ,
                    self.rightHand_frameNumber_trackedPointNumber_XYZ,
                    self.leftHand_frameNumber_trackedPointNumber_XYZ,
                    self.face_frameNumber_trackedPointNumber_XYZ,
                ]
            )
        elif len(self.body_frameNumber_trackedPointNumber_XYZ.shape) == 2:  # single frame
            return np.vstack(
                [
                    self.body_frameNumber_trackedPointNumber_XYZ,
                    self.rightHand_frameNumber_trackedPointNumber_XYZ,
                    self.leftHand_frameNumber_trackedPointNumber_XYZ,
                    self.face_frameNumber_trackedPointNumber_XYZ,
                ]
            )
        else:
            logger.error("data should have either 2 or 3 dimensions")


@dataclass
class Mediapipe2dDataPayload:
    raw_frame_payload: FramePayload = None
    mediapipe_results: Any = None
    annotated_image: np.ndarray = None
    pixel_data_numpy_arrays: Mediapipe2dNumpyArrays = None


class MediaPipeSkeletonDetector:
    def __init__(
        self,
        parameter_model=MediapipeParametersModel(),
    ):

        self._mediapipe_payload_list = []

        self._mp_drawing = mp.solutions.drawing_utils
        self._mp_drawing_styles = mp.solutions.drawing_styles
        self._mp_holistic = mp.solutions.holistic

        self._body_drawing_spec = self._mp_drawing.DrawingSpec(thickness=1, circle_radius=1)
        self._hand_drawing_spec = self._mp_drawing.DrawingSpec(thickness=1, circle_radius=1)
        self._face_drawing_spec = self._mp_drawing.DrawingSpec(thickness=1, circle_radius=1)

        self._holistic_tracker = self._mp_holistic.Holistic(
            model_complexity=parameter_model.model_complexity,
            min_detection_confidence=parameter_model.min_detection_confidence,
            min_tracking_confidence=parameter_model.min_tracking_confidence,
        )
        self._mediapipe_tracked_point_names_dict = mediapipe_tracked_point_names_dict

        self.body_names_list = self._mediapipe_tracked_point_names_dict["body"]
        self.right_hand_names_list = self._mediapipe_tracked_point_names_dict["right_hand"]
        self.left_hand_names_list = self._mediapipe_tracked_point_names_dict["left_hand"]
        self.face_names_list = self._mediapipe_tracked_point_names_dict["face"]

        # TODO - build a better iterator and list of `face_marker_names` that will only pull out the face_counters & iris edges (mp.python.solutions.face_mesh_connections.FACEMESH_CONTOURS, FACE_MESH_IRISES)

        self.number_of_body_tracked_points = len(self.body_names_list)
        self.number_of_right_hand_tracked_points = len(self.right_hand_names_list)
        self.number_of_left_hand_tracked_points = len(self.left_hand_names_list)
        self.number_of_face_tracked_points = mp.solutions.face_mesh.FACEMESH_NUM_LANDMARKS_WITH_IRISES

        self.number_of_tracked_points_total = (
            self.number_of_body_tracked_points
            + self.number_of_left_hand_tracked_points
            + self.number_of_right_hand_tracked_points
            + self.number_of_face_tracked_points
        )

    def detect_skeleton_in_image(
        self,
        raw_image: np.ndarray = None,
        annotated_image: np.ndarray = None,
    ) -> Mediapipe2dDataPayload:

        mediapipe_results = self._holistic_tracker.process(
            raw_image
        )  # <-this is where the magic happens, i.e. where the raw image is processed by a convolutional neural network to provide an estimate of joint position in pixel coordinates. Please don't forget that this is insane and should not be possible lol

        if annotated_image is None:
            annotated_image = raw_image.copy()

        annotated_image = self._annotate_image(annotated_image, mediapipe_results)

        mediapipe_single_frame_npy_data = self._list_of_mediapipe_results_to_npy_arrays(
            [mediapipe_results],
            image_width=annotated_image.shape[0],
            image_height=annotated_image.shape[1],
        )
        mediapipe_single_frame_npy_data.body_frameNumber_trackedPointNumber_XYZ = self._threshold_by_confidence(
            mediapipe_single_frame_npy_data.body_frameNumber_trackedPointNumber_XYZ,
            mediapipe_single_frame_npy_data.body_frameNumber_trackedPointNumber_confidence,
            confidence_threshold=0.5,
        )

        return Mediapipe2dDataPayload(
            mediapipe_results=mediapipe_results,
            annotated_image=annotated_image,
            pixel_data_numpy_arrays=mediapipe_single_frame_npy_data,
        )

    def process_folder_full_of_videos(
            self,
            path_to_folder_of_videos_to_process: Union[Path, str],
            output_data_folder_path: Union[str, Path],
            save_annotated_videos: bool = True,
            kill_event: multiprocessing.Event = None,
    ) -> Union[np.ndarray, None]:

        path_to_folder_of_videos_to_process = Path(path_to_folder_of_videos_to_process)

        logger.info(f"processing videos from: {path_to_folder_of_videos_to_process}")

        mediapipe2d_single_camera_npy_arrays_list = []
        for video_number, synchronized_video_file_path in enumerate(get_video_paths(path_to_folder_of_videos_to_process)):
            logger.info(f"Running `mediapipe` skeleton detection on  video: {str(synchronized_video_file_path)}")
            video_capture_object = cv2.VideoCapture(str(synchronized_video_file_path))

            video_width = video_capture_object.get(cv2.CAP_PROP_FRAME_WIDTH)
            video_height = video_capture_object.get(cv2.CAP_PROP_FRAME_HEIGHT)
            video_framerate = video_capture_object.get(cv2.CAP_PROP_FPS)

            video_mediapipe_results_list = []
            video_annotated_images_list = []

            success, image = video_capture_object.read()

            number_of_frames = int(video_capture_object.get(cv2.CAP_PROP_FRAME_COUNT))

            for frame_number in tqdm(
                range(number_of_frames),
                desc=f"mediapiping video: {synchronized_video_file_path.name}",
                total=number_of_frames,
                colour="magenta",
                unit="frames",
                dynamic_ncols=True,
            ):
                if kill_event is not None and kill_event.is_set():
                    logger.info("kill event received, exiting mediapipe process")
                    return

                if not success or image is None:
                    logger.error(f"Failed to load an image from: {str(synchronized_video_file_path)}")
                    raise Exception

                mediapipe2d_data_payload = self.detect_skeleton_in_image(raw_image=image)
                video_mediapipe_results_list.append(mediapipe2d_data_payload.mediapipe_results)
                annotated_image = self._annotate_image(image, mediapipe2d_data_payload.mediapipe_results)
                video_annotated_images_list.append(annotated_image)

                success, image = video_capture_object.read()

            if save_annotated_videos:
                annotated_video_path = path_to_folder_of_videos_to_process.parent / ANNOTATED_VIDEOS_FOLDER_NAME
                annotated_video_path.mkdir(exist_ok=True, parents=True)
                annotated_video_name = synchronized_video_file_path.stem + "_mediapipe.mp4"
                annotated_video_save_path = annotated_video_path / annotated_video_name

                video_recorder = VideoRecorder()

                logger.info(f"Saving mediapipe annotated video to : {annotated_video_save_path}")
                video_recorder.save_image_list_to_disk(
                    image_list=video_annotated_images_list,
                    path_to_save_video_file=annotated_video_save_path,
                    frames_per_second=video_framerate,
                )

            camera_mediapipe_2d_single_camera_npy_arrays = self._list_of_mediapipe_results_to_npy_arrays(
                video_mediapipe_results_list,
                image_width=video_width,
                image_height=video_height,
            )

            mediapipe2d_single_camera_npy_arrays_list.append(camera_mediapipe_2d_single_camera_npy_arrays)

        all_cameras_data2d_list = [
            m2d.all_data2d_nFrames_nTrackedPts_XY for m2d in mediapipe2d_single_camera_npy_arrays_list
        ]

        all_cameras_pose_world_data_list = [
            m2d.body_world_frameNumber_trackedPointNumber_XYZ for m2d in mediapipe2d_single_camera_npy_arrays_list
        ]

        all_cameras_right_hand_world_data_list = [
            m2d.rightHand_frameNumber_trackedPointNumber_XYZ for m2d in mediapipe2d_single_camera_npy_arrays_list
        ]

        all_cameras_left_hand_world_data_list = [
            m2d.leftHand_frameNumber_trackedPointNumber_XYZ for m2d in mediapipe2d_single_camera_npy_arrays_list
        ]

        all_cameras_face_world_data_list = [
            m2d.face_frameNumber_trackedPointNumber_XYZ for m2d in mediapipe2d_single_camera_npy_arrays_list
        ]

        number_of_cameras = len(all_cameras_data2d_list)
        number_of_frames = all_cameras_data2d_list[0].shape[0]
        number_of_tracked_points = all_cameras_data2d_list[0].shape[1]
        number_of_spatial_dimensions = all_cameras_data2d_list[0].shape[2]

        number_of_body_points = all_cameras_pose_world_data_list[0].shape[1]

        data2d_numCams_numFrames_numTrackedPts_XY = np.empty(
            (
                number_of_cameras,
                number_of_frames,
                number_of_tracked_points,
                number_of_spatial_dimensions,
            )
        )

        body_world_numCams_numFrames_numTrackedPts_XYZ = np.empty(
            (
                number_of_cameras,
                number_of_frames,
                number_of_tracked_points,
                number_of_spatial_dimensions,
            )
        )

        for cam_num in range(number_of_cameras):
            data2d_numCams_numFrames_numTrackedPts_XY[cam_num, :, :, :] = all_cameras_data2d_list[cam_num]

            pose_3d = all_cameras_pose_world_data_list[cam_num]
            right_hand_3d = all_cameras_right_hand_world_data_list[cam_num]
            left_hand_3d = all_cameras_left_hand_world_data_list[cam_num]
            face_3d = all_cameras_face_world_data_list[cam_num]

            body_world_numCams_numFrames_numTrackedPts_XYZ[cam_num, :, :, :] = np.concatenate((
                pose_3d,
                right_hand_3d,
                left_hand_3d,
                face_3d),
                axis=1)

            logger.info(
                f"The shape of body_world_numCams_numFrames_numTrackedPts_XYZ is "
                f"{body_world_numCams_numFrames_numTrackedPts_XYZ.shape}")

        self._save_mediapipe2d_data_to_npy(
            data2d_numCams_numFrames_numTrackedPts_XY=data2d_numCams_numFrames_numTrackedPts_XY,
            body_world_numCams_numFrames_numTrackedPts_XYZ=body_world_numCams_numFrames_numTrackedPts_XYZ,
            output_data_folder_path=Path(output_data_folder_path),
        )
        return data2d_numCams_numFrames_numTrackedPts_XY

    def _save_mediapipe2d_data_to_npy(
        self,
        data2d_numCams_numFrames_numTrackedPts_XY: np.ndarray,
        body_world_numCams_numFrames_numTrackedPts_XYZ: np.ndarray,
        output_data_folder_path: Union[str, Path],
    ):
        mediapipe_2dData_save_path = Path(output_data_folder_path) / MEDIAPIPE_2D_NPY_FILE_NAME
        mediapipe_2dData_save_path.parent.mkdir(exist_ok=True, parents=True)
        logger.info(f"saving mediapipe image npy file: {mediapipe_2dData_save_path}")
        np.save(str(mediapipe_2dData_save_path), data2d_numCams_numFrames_numTrackedPts_XY)

        mediapipe_body_world_save_path = Path(output_data_folder_path) / MEDIAPIPE_BODY_WORLD_FILE_NAME
        mediapipe_body_world_save_path.parent.mkdir(exist_ok=True, parents=True)
        logger.info(f"saving mediapipe body world npy xyz: {mediapipe_body_world_save_path}")
        np.save(str(mediapipe_body_world_save_path), body_world_numCams_numFrames_numTrackedPts_XYZ)

    def _annotate_image(self, image, mediapipe_results):
        self._mp_drawing.draw_landmarks(
            image=image,
            landmark_list=mediapipe_results.face_landmarks,
            connections=self._mp_holistic.FACEMESH_CONTOURS,
            landmark_drawing_spec=None,
            connection_drawing_spec=self._mp_drawing_styles.get_default_face_mesh_contours_style(),
        )
        self._mp_drawing.draw_landmarks(
            image=image,
            landmark_list=mediapipe_results.face_landmarks,
            connections=self._mp_holistic.FACEMESH_TESSELATION,
            landmark_drawing_spec=None,
            connection_drawing_spec=self._mp_drawing_styles.get_default_face_mesh_tesselation_style(),
        )
        self._mp_drawing.draw_landmarks(
            image=image,
            landmark_list=mediapipe_results.pose_landmarks,
            connections=self._mp_holistic.POSE_CONNECTIONS,
            landmark_drawing_spec=self._mp_drawing_styles.get_default_pose_landmarks_style(),
        )
        self._mp_drawing.draw_landmarks(
            image=image,
            landmark_list=mediapipe_results.left_hand_landmarks,
            connections=self._mp_holistic.HAND_CONNECTIONS,
            landmark_drawing_spec=None,
            connection_drawing_spec=self._mp_drawing_styles.get_default_hand_connections_style(),
        )

        self._mp_drawing.draw_landmarks(
            image=image,
            landmark_list=mediapipe_results.right_hand_landmarks,
            connections=self._mp_holistic.HAND_CONNECTIONS,
            landmark_drawing_spec=None,
            connection_drawing_spec=self._mp_drawing_styles.get_default_hand_connections_style(),
        )
        return image

    def _list_of_mediapipe_results_to_npy_arrays(
        self,
        mediapipe_results_list: List,
        image_width: Union[int, float],
        image_height: Union[int, float],
    ) -> Mediapipe2dNumpyArrays:

        number_of_frames = len(mediapipe_results_list)
        number_of_spatial_dimensions = 3  # this will be 2d XY pixel data, with mediapipe's estimate of Z

        body_frameNumber_trackedPointNumber_XYZ = np.zeros(
            (
                number_of_frames,
                self.number_of_body_tracked_points,
                number_of_spatial_dimensions,
            )
        )
        body_frameNumber_trackedPointNumber_XYZ[:] = np.nan

        body_world_frameNumber_trackedPointNumber_XYZ = np.zeros(
            (
                number_of_frames,
                self.number_of_body_tracked_points,
                number_of_spatial_dimensions,
            )
        )
        body_world_frameNumber_trackedPointNumber_XYZ[:] = np.nan

        body_frameNumber_trackedPointNumber_confidence = np.zeros(
            (number_of_frames, self.number_of_body_tracked_points)
        )
        body_frameNumber_trackedPointNumber_confidence[:] = np.nan  # only body markers get a 'confidence' value

        rightHand_frameNumber_trackedPointNumber_XYZ = np.zeros(
            (
                number_of_frames,
                self.number_of_right_hand_tracked_points,
                number_of_spatial_dimensions,
            )
        )
        rightHand_frameNumber_trackedPointNumber_XYZ[:] = np.nan

        leftHand_frameNumber_trackedPointNumber_XYZ = np.zeros(
            (
                number_of_frames,
                self.number_of_left_hand_tracked_points,
                number_of_spatial_dimensions,
            )
        )
        leftHand_frameNumber_trackedPointNumber_XYZ[:] = np.nan

        face_frameNumber_trackedPointNumber_XYZ = np.zeros(
            (
                number_of_frames,
                self.number_of_face_tracked_points,
                number_of_spatial_dimensions,
            )
        )
        face_frameNumber_trackedPointNumber_XYZ[:] = np.nan

        all_body_tracked_points_visible_on_frame_bool_list = []
        all_right_hand_points_visible_on_frame_bool_list = []
        all_left_hand_points_visible_on_frame_bool_list = []
        all_face_points_visible_on_frame_bool_list = []
        all_tracked_points_visible_on_frame_list = []

        for frame_number, frame_results in enumerate(mediapipe_results_list):

            # get the Body data (aka 'pose')
            if frame_results.pose_landmarks is not None:

                for landmark_number, landmark_data in enumerate(frame_results.pose_landmarks.landmark):
                    body_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 0] = (
                        landmark_data.x * image_width
                    )
                    body_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 1] = (
                        landmark_data.y * image_height
                    )
                    body_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 2] = (
                        landmark_data.z * image_width  # z is on roughly the same scale as x, according to mediapipe docs
                    )
                    body_frameNumber_trackedPointNumber_confidence[
                        frame_number, landmark_number
                    ] = landmark_data.visibility  # mediapipe calls their 'confidence' score 'visibility'

                for landmark_number, landmark_data in enumerate(frame_results.pose_world_landmarks.landmark):
                    body_world_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 0] = (
                            landmark_data.x * image_width
                    )
                    body_world_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 1] = (
                            landmark_data.y * image_height
                    )
                    body_world_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 2] = (
                            landmark_data.z * image_width
                    )

            # get Right Hand data
            if frame_results.right_hand_landmarks is not None:
                for landmark_number, landmark_data in enumerate(frame_results.right_hand_landmarks.landmark):
                    rightHand_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 0] = (
                        landmark_data.x * image_width
                    )
                    rightHand_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 1] = (
                        landmark_data.y * image_height
                    )
                    rightHand_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 2] = (
                            landmark_data.z * image_width
                    )

            # get Left Hand data
            if frame_results.left_hand_landmarks is not None:
                for landmark_number, landmark_data in enumerate(frame_results.left_hand_landmarks.landmark):
                    leftHand_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 0] = (
                        landmark_data.x * image_width
                    )
                    leftHand_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 1] = (
                        landmark_data.y * image_height
                    )
                    leftHand_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 2] = (
                            landmark_data.z * image_width
                    )

            # get Face data
            if frame_results.face_landmarks is not None:
                for landmark_number, landmark_data in enumerate(frame_results.face_landmarks.landmark):
                    face_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 0] = (
                        landmark_data.x * image_width
                    )
                    face_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 1] = (
                        landmark_data.y * image_height
                    )
                    face_frameNumber_trackedPointNumber_XYZ[frame_number, landmark_number, 2] = (
                            landmark_data.z * image_width
                    )

            # check if all tracked points are visible on this frame
            all_body_visible = all(sum(np.isnan(body_frameNumber_trackedPointNumber_XYZ[frame_number, :, :])) == 0)
            all_body_tracked_points_visible_on_frame_bool_list.append(all_body_visible)

            all_right_hand_visible = all(
                sum(np.isnan(rightHand_frameNumber_trackedPointNumber_XYZ[frame_number, :, :])) == 0
            )
            all_right_hand_points_visible_on_frame_bool_list.append(all_right_hand_visible)

            all_left_hand_visible = all(
                sum(np.isnan(leftHand_frameNumber_trackedPointNumber_XYZ[frame_number, :, :])) == 0
            )
            all_left_hand_points_visible_on_frame_bool_list.append(all_left_hand_visible)

            all_face_visible = all(sum(np.isnan(face_frameNumber_trackedPointNumber_XYZ[frame_number, :, :])) == 0)
            all_face_points_visible_on_frame_bool_list.append(all_face_visible)

            all_points_visible = all(
                [
                    all_body_visible,
                    all_right_hand_visible,
                    all_left_hand_visible,
                    all_face_visible,
                ],
            )

            all_tracked_points_visible_on_frame_list.append(all_points_visible)

        return Mediapipe2dNumpyArrays(
            body_frameNumber_trackedPointNumber_XYZ=np.squeeze(body_frameNumber_trackedPointNumber_XYZ),
            body_world_frameNumber_trackedPointNumber_XYZ=np.squeeze(body_world_frameNumber_trackedPointNumber_XYZ),
            rightHand_frameNumber_trackedPointNumber_XYZ=np.squeeze(rightHand_frameNumber_trackedPointNumber_XYZ),
            leftHand_frameNumber_trackedPointNumber_XYZ=np.squeeze(leftHand_frameNumber_trackedPointNumber_XYZ),
            face_frameNumber_trackedPointNumber_XYZ=np.squeeze(face_frameNumber_trackedPointNumber_XYZ),
            body_frameNumber_trackedPointNumber_confidence=np.squeeze(
                body_frameNumber_trackedPointNumber_confidence
            ),
        )

    def _threshold_by_confidence(
        self,
        data2d_trackedPoint_dim: np.ndarray,
        data2d_trackedPoint_confidence: np.ndarray,
        confidence_threshold: float,
    ):

        threshold_mask = data2d_trackedPoint_confidence < confidence_threshold
        data2d_trackedPoint_dim[threshold_mask, :] = np.nan
        return data2d_trackedPoint_dim



import logging
import threading
from pathlib import Path
from typing import Callable

from PyQt6.QtCore import pyqtSignal, pyqtSlot, Qt
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QPushButton, QVBoxLayout, QWidget, QGroupBox
from pyqtgraph.parametertree import Parameter, ParameterTree

from freemocap.gui.qt.widgets.control_panel.calibration_control_panel import CalibrationControlPanel
from freemocap.gui.qt.widgets.control_panel.process_mocap_data_panel.parameter_groups.create_parameter_groups import (
    create_mediapipe_parameter_group, create_3d_triangulation_prarameter_group, create_post_processing_parameter_group,
)
from freemocap.gui.qt.workers.process_motion_capture_data_thread_worker import (
    ProcessMotionCaptureDataThreadWorker,
)
from freemocap.parameter_info_models.recording_processing_parameter_models import (
    RecordingProcessingParameterModel,
)

logger = logging.getLogger(__name__)


class ProcessMotionCaptureDataPanel(QWidget):
    processing_finished_signal = pyqtSignal()

    def __init__(
            self,
            recording_processing_parameters: RecordingProcessingParameterModel,
            get_active_recording_info: Callable,
            kill_thread_event: threading.Event,
            parent=None,
    ):
        super().__init__(parent=parent)

        self._kill_thread_event = kill_thread_event
        self._get_active_recording_info = get_active_recording_info
        self._recording_processing_parameter_model = recording_processing_parameters
        self._recording_processing_parameter_model.recording_info_model = self._get_active_recording_info()

        self._layout = QVBoxLayout()
        self._layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setLayout(self._layout)

        calibration_group_box = QGroupBox("Selected Capture Volume Calibration TOML:")
        vbox = QVBoxLayout()
        calibration_group_box.setLayout(vbox)
        self._layout.addWidget(calibration_group_box)

        self._calibration_control_panel = CalibrationControlPanel(
            get_active_recording_info_callable=self._get_active_recording_info,
            kill_thread_event=self._kill_thread_event,
            parent=self,
        )
        vbox.addWidget(self._calibration_control_panel)

        self._process_motion_capture_data_button = QPushButton(
            "Process Motion Capture Videos",
        )
        self._process_motion_capture_data_button.setFont(QFont("Dosis", 16, QFont.Weight.Bold))
        
        self._process_motion_capture_data_button.clicked.connect(self._launch_process_motion_capture_data_thread_worker)
        self._layout.addWidget(self._process_motion_capture_data_button)

        self._parameter_tree_widget = ParameterTree(parent=self, showHeader=False)
        self._layout.addWidget(self._parameter_tree_widget)

        self._add_parameters_to_parameter_tree_widget(
            self._parameter_tree_widget,
            session_processing_parameter_model=self._recording_processing_parameter_model,
        )

        self._process_motion_capture_data_thread_worker = None

    @property
    def process_motion_capture_data_button(self):
        return self._process_motion_capture_data_button

    def calibrate_from_active_recording(self, charuco_square_size_mm: float):
        self._calibration_control_panel.calibrate_from_active_recording(charuco_square_size_mm=charuco_square_size_mm)

    def update_calibration_path(self) -> bool:
        self._calibration_control_panel.update_calibrate_from_active_recording_button_text()
        return self._calibration_control_panel.update_calibration_toml_path()

    def _add_parameters_to_parameter_tree_widget(
            self,
            parameter_tree_widget: ParameterTree,
            session_processing_parameter_model: RecordingProcessingParameterModel,
    ):
        parameter_group = self._convert_session_processing_parameter_model_to_parameter_group(
            session_processing_parameter_model
        )
        parameter_tree_widget.setParameters(parameter_group, showTop=False)
        parameter_tree_widget.setObjectName("parameter-tree-widget")
        return parameter_group

    def _convert_session_processing_parameter_model_to_parameter_group(
            self,
            session_processing_parameter_model: RecordingProcessingParameterModel,
    ):

        return Parameter.create(
            name="Processing Parameters",
            type="group",
            children=[
                dict(
                    name="2d Image Trackers",
                    type="group",
                    children=[
                        self._create_new_skip_this_step_parameter(),
                        create_mediapipe_parameter_group(session_processing_parameter_model.mediapipe_parameters_model),
                    ],
                    tip="Methods for tracking 2d points in images (e.g. mediapipe, deeplabcut(TODO), openpose(TODO), etc ...)",
                ),
                dict(
                    name="3d triangulation methods",
                    type="group",
                    children=[
                        self._create_new_skip_this_step_parameter(),
                        create_3d_triangulation_prarameter_group(
                            session_processing_parameter_model.anipose_triangulate_3d_parameters_model
                        ),
                    ],
                    tip="Methods for triangulating 3d points from 2d points (using epipolar geometry and the 'camera_calibration' data).",
                ),
                dict(
                    name="Post Processing (data cleaning)",
                    type="group",
                    children=[
                        self._create_new_skip_this_step_parameter(),
                        create_post_processing_parameter_group(
                            session_processing_parameter_model.post_processing_parameters_model
                        ),
                    ],
                    tip="Methods for cleaning up the data (e.g. filtering/smoothing, gap filling, etc ...)"
                        "TODO - Add/expose more post processing methods here (e.g. gap filling, outlier removal, etc ...)",
                ),
            ],
        )

    def _extract_session_parameter_model_from_parameter_tree(
            self,
    ) -> RecordingProcessingParameterModel:
        # rec = extract_mediapipe_parameter_model_from_parameter_tree(self._parameter_tree_widget)
        logger.debug("TODO - extract session parameter model from parameter tree")
        rec = self._recording_processing_parameter_model
        rec.recording_info_model = self._get_active_recording_info()

        if self._calibration_control_panel.calibration_toml_path:
            rec.recording_info_model.calibration_toml_path = self._calibration_control_panel.calibration_toml_path
        else:
            rec.recording_info_model.calibration_toml_path = self._calibration_control_panel.open_load_camera_calibration_toml_dialog()

        if not rec.recording_info_model.calibration_toml_path:
            logger.error(
                "No calibration TOML selected - Processing will fail at the '3d triangulation' step (but it will get through the 'Tracking things in 2d images' step).")

        return rec

    def _create_new_skip_this_step_parameter(self):
        parameter = Parameter.create(
            name="Skip this step?",
            type="bool",
            value=False,
            tip="If you have already run this step, you can skip it." "re-running it will overwrite the existing data.",
        )
        parameter.sigValueChanged.connect(self.disable_this_parameter_group)

        return parameter

    def disable_this_parameter_group(self, parameter, changes):
        logger.debug(f"TODO - disable parameter group when 'skip this step?' is checked")
        pass
        # skip_this_step_bool = parameter.value()
        # parameter_group = parameter.parent()
        # for child in parameter_group.children():
        #     if child.name() != "Skip this step?":
        #         logger.debug(f"Disabling {child.name()}")
        #         child.setOpts(enabled=not skip_this_step_bool)

    def _launch_process_motion_capture_data_thread_worker(self):
        logger.debug("Launching process motion capture data process")
        session_parameter_model = self._extract_session_parameter_model_from_parameter_tree()
        session_parameter_model.recording_info_model = self._get_active_recording_info()

        if session_parameter_model.recording_info_model is None:
            logger.error("No active recording selected.")
            return
        if not Path(session_parameter_model.recording_info_model.path).exists():
            logger.error(f"Recording path does not exist: {session_parameter_model.recording_info_model.path}.")
            return

        session_parameter_model.recording_info_model.calibration_toml_path = self._calibration_control_panel.calibration_toml_path

        self._process_motion_capture_data_thread_worker = ProcessMotionCaptureDataThreadWorker(session_parameter_model,
                                                                                               kill_event=self._kill_thread_event)
        self._process_motion_capture_data_thread_worker.start()

        self._process_motion_capture_data_thread_worker.finished.connect(self._handle_finished_signal)

    @pyqtSlot()
    def _handle_finished_signal(self):
        logger.debug("Process motion capture data process finished.")
        self.processing_finished_signal.emit()


if __name__ == "__main__":
    from PyQt6.QtWidgets import QApplication
    import sys

    app = QApplication(sys.argv)
    window = ProcessMotionCaptureDataPanel()
    window.show()
    app.exec()



import logging
from pathlib import Path
from typing import Union

import cv2

from freemocap.utilities.get_video_paths import get_video_paths

logger = logging.getLogger(__name__)


def get_number_of_frames_of_videos_in_a_folder(folder_path: Union[str, Path]):
    """
    Get the number of frames in the first video in a folder
    """

    list_of_video_paths = get_video_paths(Path(folder_path))

    if len(list_of_video_paths) == 0:
        logger.error(f"No videos found in {folder_path}")
        return None

    frame_count = []
    for video_path in list_of_video_paths:
        cap = cv2.VideoCapture(str(video_path))
        frame_count.append(int(cap.get(cv2.CAP_PROP_FRAME_COUNT)))
        cap.release()


    return frame_count



from pathlib import Path
from typing import Union

import numpy as np

from freemocap.tests.utilities.get_number_of_frames_of_videos_in_a_folder import (
    get_number_of_frames_of_videos_in_a_folder,
)
from freemocap.utilities.get_video_paths import get_video_paths


def test_mediapipe_2d_data_shape(
    synchronized_videos_folder: Union[str, Path],
    mediapipe_2d_data_file_path: Union[str, Path],
):

    """
    test that the `mediapipe 2d detection` process worked correctly by checking:

    1. There is an `.npy` file containing the mediapipe 2d data in the `output_data_folder`
    2. The dimensions of that `npy` (number of cameras, number of frames, [ need to do - number of tracked points], [pixelX, pixelY] matches the videos in the `synchronized videos` folder

    TODO - check number of tracked points vs 'expected' number of tracked points
    """

    assert Path(
        mediapipe_2d_data_file_path
    ).is_file(), f"{mediapipe_2d_data_file_path} is not a file"

    mediapipe_2d_data = np.load(mediapipe_2d_data_file_path)

    list_of_video_paths = get_video_paths(Path(synchronized_videos_folder))
    number_of_videos = len(list_of_video_paths)
    assert mediapipe_2d_data.shape[0] == number_of_videos

    frame_count = get_number_of_frames_of_videos_in_a_folder(synchronized_videos_folder)

    assert (
        len(set(frame_count)) == 1
    ), f"Videos in {synchronized_videos_folder} have different frame counts: {frame_count}"
    assert (
        mediapipe_2d_data.shape[1] == frame_count[0]
    ), f"Number of frames in {mediapipe_2d_data_file_path} does not match number of frames of videos in {synchronized_videos_folder}"

    # TODO - check number of tracked points vs 'expected' number of tracked points

    assert mediapipe_2d_data.shape[3] == 3  # x y z

    return True



# %%
import logging
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
from mediapipe.python.solutions import holistic as mp_holistic

logger = logging.getLogger(__name__)


def convert_mediapipe_npy_to_csv(
    mediapipe_3d_frame_trackedPoint_xyz: np.ndarray,
    output_data_folder_path: Union[str, Path],
):
    logger.info(
        f"Converting npy data with shape: {mediapipe_3d_frame_trackedPoint_xyz.shape} into `csv` and smaller `npy` files"
    )

    # %%
    mediapipe_pose_landmark_names = [
        landmark.name.lower() for landmark in mp_holistic.PoseLandmark
    ]
    mediapipe_hand_landmark_names = [
        landmark.name.lower() for landmark in mp_holistic.HandLandmark
    ]
    # face_landmark_names = [landmark.name.lower() for landmark in mp_holistic.PoseLandmark] #gonna have the clever for the face
    # logger.info(f"Body tracked point names: {mediapipe_pose_landmark_names}")
    # logger.info(mediapipe_hand_landmark_names)

    # %%
    # get number of points in body, hands, face

    number_of_body_points = len(mediapipe_pose_landmark_names)
    number_of_hand_points = len(mediapipe_hand_landmark_names)

    first_body_marker_index = 0
    last_body_marker_index = number_of_body_points - 1

    first_right_hand_marker_index = last_body_marker_index + 1
    last_right_hand_marker_index = number_of_body_points + number_of_hand_points - 1

    first_left_hand_marker_index = last_right_hand_marker_index + 1
    last_left_hand_marker_index = (
        last_right_hand_marker_index + 1 + number_of_hand_points - 1
    )

    first_face_marker_index = last_left_hand_marker_index + 1
    last_face_marker_index = mediapipe_3d_frame_trackedPoint_xyz.shape[1]

    number_of_face_points = last_face_marker_index - first_face_marker_index

    # logger.info(
    #     f"body tracked point indices: {first_body_marker_index}:{last_body_marker_index}"
    # )
    # logger.info(
    #     f"right hand tracked point indices: {first_right_hand_marker_index}:{last_right_hand_marker_index}"
    # )
    # logger.info(
    #     f"left hand tracked point indices: {first_left_hand_marker_index}:{last_left_hand_marker_index}"
    # )
    # logger.info(
    #     f"face tracked point indices: {first_face_marker_index}:{last_face_marker_index}"
    # )
    #
    # logger.info(
    #     f"number of body points: {last_body_marker_index - first_body_marker_index + 1}"
    # )
    # logger.info(
    #     f"number of right hand points: {last_right_hand_marker_index - first_right_hand_marker_index + 1}"
    # )
    # logger.info(
    #     f"number of left hand points: {last_left_hand_marker_index - first_left_hand_marker_index + 1}"
    # )
    # logger.info(
    #     f"number of face points: {last_face_marker_index - first_face_marker_index + 1}"
    # )

    # %%
    body_3d_xyz = mediapipe_3d_frame_trackedPoint_xyz[
        :, first_body_marker_index : last_body_marker_index + 1, :
    ]
    right_hand_3d_xyz = mediapipe_3d_frame_trackedPoint_xyz[
        :, first_right_hand_marker_index : last_right_hand_marker_index + 1, :
    ]
    left_hand_3d_xyz = mediapipe_3d_frame_trackedPoint_xyz[
        :, first_left_hand_marker_index : last_left_hand_marker_index + 1, :
    ]
    face_3d_xyz = mediapipe_3d_frame_trackedPoint_xyz[
        :, first_face_marker_index : last_face_marker_index + 1, :
    ]

    # logger.info(f"body 3d xyz shape: {body_3d_xyz.shape}")
    # logger.info(f"right hand 3d xyz shape: {right_hand_3d_xyz.shape}")
    # logger.info(f"left hand 3d xyz shape: {left_hand_3d_xyz.shape}")
    # logger.info(f"face 3d xyz shape: {face_3d_xyz.shape}")

    # %%
    # save broken up npy files
    np.save(
        str(Path(output_data_folder_path) / "mediapipe_body_3d_xyz.npy"), body_3d_xyz
    )
    np.save(
        str(Path(output_data_folder_path) / "mediapipe_right_hand_3d_xyz.npy"),
        right_hand_3d_xyz,
    )
    np.save(
        str(Path(output_data_folder_path) / "mediapipe_left_hand_3d_xyz.npy"),
        left_hand_3d_xyz,
    )
    np.save(
        str(Path(output_data_folder_path) / "mediapipe_face_3d_xyz.npy"), face_3d_xyz
    )

    # %%
    # create pandas data frame headers

    body_3d_xyz_header = []
    for landmark_name in mediapipe_pose_landmark_names:
        body_3d_xyz_header.append(f"{landmark_name}_x")
        body_3d_xyz_header.append(f"{landmark_name}_y")
        body_3d_xyz_header.append(f"{landmark_name}_z")

    right_hand_3d_xyz_header = []
    for landmark_name in mediapipe_hand_landmark_names:
        right_hand_3d_xyz_header.append(f"right_hand_{landmark_name}_x")
        right_hand_3d_xyz_header.append(f"right_hand_{landmark_name}_y")
        right_hand_3d_xyz_header.append(f"right_hand_{landmark_name}_z")

    left_hand_3d_xyz_header = []
    for landmark_name in mediapipe_hand_landmark_names:
        left_hand_3d_xyz_header.append(f"left_hand_{landmark_name}_x")
        left_hand_3d_xyz_header.append(f"left_hand_{landmark_name}_y")
        left_hand_3d_xyz_header.append(f"left_hand_{landmark_name}_z")

    face_3d_xyz_header = []
    for landmark_number in range(last_face_marker_index - first_face_marker_index):
        face_3d_xyz_header.append(f"face_{str(landmark_number).zfill(4)}_x")
        face_3d_xyz_header.append(f"face_{str(landmark_number).zfill(4)}_y")
        face_3d_xyz_header.append(f"face_{str(landmark_number).zfill(4)}_z")
    #
    # logger.debug(
    #     f"length of body 3d xyz header: {len(body_3d_xyz_header)}, should be: {number_of_body_points * 3}"
    # )
    # logger.debug(
    #     f"length of right hand 3d xyz header: {len(right_hand_3d_xyz_header)}, should be: {number_of_hand_points * 3}"
    # )
    # logger.debug(
    #     f"length of left hand 3d xyz header: {len(left_hand_3d_xyz_header)}, should be: {number_of_hand_points * 3}"
    # )
    # logger.debug(
    #     f"length of face 3d xyz header: {len(face_3d_xyz_header)}, should be: {number_of_face_points * 3}"
    # )

    # %%
    number_of_frames = mediapipe_3d_frame_trackedPoint_xyz.shape[0]
    body_flat = body_3d_xyz.reshape(number_of_frames, number_of_body_points * 3)

    body_dataframe = pd.DataFrame(body_flat, columns=body_3d_xyz_header)
    body_dataframe.to_csv(
        str(Path(output_data_folder_path) / "mediapipe_body_3d_xyz.csv"), index=False
    )

    right_hand_flat = right_hand_3d_xyz.reshape(
        number_of_frames, number_of_hand_points * 3
    )
    right_hand_dataframe = pd.DataFrame(
        right_hand_flat, columns=right_hand_3d_xyz_header
    )
    right_hand_dataframe.to_csv(
        str(Path(output_data_folder_path) / "mediapipe_right_hand_3d_xyz.csv"),
        index=False,
    )

    left_hand_flat = left_hand_3d_xyz.reshape(
        number_of_frames, number_of_hand_points * 3
    )
    left_hand_dataframe = pd.DataFrame(left_hand_flat, columns=left_hand_3d_xyz_header)
    left_hand_dataframe.to_csv(
        str(Path(output_data_folder_path) / "mediapipe_left_hand_3d_xyz.csv"),
        index=False,
    )

    face_flat = face_3d_xyz.reshape(number_of_frames, number_of_face_points * 3)
    face_dataframe = pd.DataFrame(face_flat, columns=face_3d_xyz_header)
    face_dataframe.to_csv(
        str(Path(output_data_folder_path) / "mediapipe_face_3d_xyz.csv"), index=False
    )

    # %%

    # %%
    mediapipe_pose_landmark_names = [
        landmark.name.lower() for landmark in mp_holistic.PoseLandmark
    ]
    mediapipe_hand_landmark_names = [
        landmark.name.lower() for landmark in mp_holistic.HandLandmark
    ]
    # face_landmark_names = [landmark.name.lower() for landmark in mp_holistic.PoseLandmark] #gonna have the clever for the face

    # logger.debug(f"Body tracked point names: {mediapipe_pose_landmark_names}")
    # logger.debug(mediapipe_hand_landmark_names)

    # %%
    # get number of points in body, hands, face

    number_of_body_points = len(mediapipe_pose_landmark_names)
    number_of_hand_points = len(mediapipe_hand_landmark_names)

    first_body_marker_index = 0
    last_body_marker_index = number_of_body_points - 1

    first_right_hand_marker_index = last_body_marker_index + 1
    last_right_hand_marker_index = number_of_body_points + number_of_hand_points - 1

    first_left_hand_marker_index = last_right_hand_marker_index + 1
    last_left_hand_marker_index = (
        last_right_hand_marker_index + 1 + number_of_hand_points - 1
    )

    first_face_marker_index = last_left_hand_marker_index + 1
    last_face_marker_index = mediapipe_3d_frame_trackedPoint_xyz.shape[1]

    number_of_face_points = last_face_marker_index - first_face_marker_index

    # logger.debug(
    #     f"body tracked point indices: {first_body_marker_index}:{last_body_marker_index}"
    # )
    # logger.debug(
    #     f"right hand tracked point indices: {first_right_hand_marker_index}:{last_right_hand_marker_index}"
    # )
    # logger.debug(
    #     f"left hand tracked point indices: {first_left_hand_marker_index}:{last_left_hand_marker_index}"
    # )
    # logger.debug(
    #     f"face tracked point indices: {first_face_marker_index}:{last_face_marker_index}"
    # )
    #
    # logger.debug(
    #     f"number of body points: {last_body_marker_index - first_body_marker_index + 1}"
    # )
    # logger.debug(
    #     f"number of right hand points: {last_right_hand_marker_index - first_right_hand_marker_index + 1}"
    # )
    # logger.debug(
    #     f"number of left hand points: {last_left_hand_marker_index - first_left_hand_marker_index + 1}"
    # )
    # logger.debug(
    #     f"number of face points: {last_face_marker_index - first_face_marker_index + 1}"
    # )

    # %%
    body_3d_xyz = mediapipe_3d_frame_trackedPoint_xyz[
        :, first_body_marker_index : last_body_marker_index + 1, :
    ]
    right_hand_3d_xyz = mediapipe_3d_frame_trackedPoint_xyz[
        :, first_right_hand_marker_index : last_right_hand_marker_index + 1, :
    ]
    left_hand_3d_xyz = mediapipe_3d_frame_trackedPoint_xyz[
        :, first_left_hand_marker_index : last_left_hand_marker_index + 1, :
    ]
    face_3d_xyz = mediapipe_3d_frame_trackedPoint_xyz[
        :, first_face_marker_index : last_face_marker_index + 1, :
    ]

    logger.debug(f"body 3d xyz shape: {body_3d_xyz.shape}")
    logger.debug(f"right hand 3d xyz shape: {right_hand_3d_xyz.shape}")
    logger.debug(f"left hand 3d xyz shape: {left_hand_3d_xyz.shape}")
    logger.debug(f"face 3d xyz shape: {face_3d_xyz.shape}")

    # %%
    # save broken up npy files
    np.save(
        str(Path(output_data_folder_path) / "mediapipe_body_3d_xyz.npy"), body_3d_xyz
    )
    np.save(
        str(Path(output_data_folder_path) / "mediapipe_right_hand_3d_xyz.npy"),
        right_hand_3d_xyz,
    )
    np.save(
        str(Path(output_data_folder_path) / "mediapipe_left_hand_3d_xyz.npy"),
        left_hand_3d_xyz,
    )
    np.save(
        str(Path(output_data_folder_path) / "mediapipe_face_3d_xyz.npy"), face_3d_xyz
    )

    # %%
    # create pandas data frame headers

    body_3d_xyz_header = []
    for landmark_name in mediapipe_pose_landmark_names:
        body_3d_xyz_header.append(f"{landmark_name}_x")
        body_3d_xyz_header.append(f"{landmark_name}_y")
        body_3d_xyz_header.append(f"{landmark_name}_z")

    right_hand_3d_xyz_header = []
    for landmark_name in mediapipe_hand_landmark_names:
        right_hand_3d_xyz_header.append(f"right_hand_{landmark_name}_x")
        right_hand_3d_xyz_header.append(f"right_hand_{landmark_name}_y")
        right_hand_3d_xyz_header.append(f"right_hand_{landmark_name}_z")

    left_hand_3d_xyz_header = []
    for landmark_name in mediapipe_hand_landmark_names:
        left_hand_3d_xyz_header.append(f"left_hand_{landmark_name}_x")
        left_hand_3d_xyz_header.append(f"left_hand_{landmark_name}_y")
        left_hand_3d_xyz_header.append(f"left_hand_{landmark_name}_z")

    face_3d_xyz_header = []
    for landmark_number in range(last_face_marker_index - first_face_marker_index):
        face_3d_xyz_header.append(f"face_{str(landmark_number).zfill(4)}_x")
        face_3d_xyz_header.append(f"face_{str(landmark_number).zfill(4)}_y")
        face_3d_xyz_header.append(f"face_{str(landmark_number).zfill(4)}_z")

    logger.debug(
        f"length of body 3d xyz `.csv` header: {len(body_3d_xyz_header)}, should be: {number_of_body_points * 3}"
    )
    logger.debug(
        f"length of right hand 3d xyz `.csv` header: {len(right_hand_3d_xyz_header)}, should be: {number_of_hand_points * 3}"
    )
    logger.debug(
        f"length of left hand 3d xyz `.csv` header: {len(left_hand_3d_xyz_header)}, should be: {number_of_hand_points * 3}"
    )
    logger.debug(
        f"length of face 3d xyz `.csv` header: {len(face_3d_xyz_header)}, should be: {number_of_face_points * 3}"
    )

    # %%
    number_of_frames = mediapipe_3d_frame_trackedPoint_xyz.shape[0]
    body_flat = body_3d_xyz.reshape(number_of_frames, number_of_body_points * 3)

    body_dataframe = pd.DataFrame(body_flat, columns=body_3d_xyz_header)
    body_dataframe.to_csv(
        str(Path(output_data_folder_path) / "mediapipe_body_3d_xyz.csv"), index=False
    )

    right_hand_flat = right_hand_3d_xyz.reshape(
        number_of_frames, number_of_hand_points * 3
    )
    right_hand_dataframe = pd.DataFrame(
        right_hand_flat, columns=right_hand_3d_xyz_header
    )
    right_hand_dataframe.to_csv(
        str(Path(output_data_folder_path) / "mediapipe_right_hand_3d_xyz.csv"),
        index=False,
    )

    left_hand_flat = left_hand_3d_xyz.reshape(
        number_of_frames, number_of_hand_points * 3
    )
    left_hand_dataframe = pd.DataFrame(left_hand_flat, columns=left_hand_3d_xyz_header)
    left_hand_dataframe.to_csv(
        str(Path(output_data_folder_path) / "mediapipe_left_hand_3d_xyz.csv"),
        index=False,
    )

    face_flat = face_3d_xyz.reshape(number_of_frames, number_of_face_points * 3)
    face_dataframe = pd.DataFrame(face_flat, columns=face_3d_xyz_header)
    face_dataframe.to_csv(
        str(Path(output_data_folder_path) / "mediapipe_face_3d_xyz.csv"), index=False
    )

    logger.info("Done saving out `csv` and broken up `npy` files")


if __name__ == "__main__":
    mediapipe_3d_frame_trackedPoint_xyz = np.load(
        r"C:\Users\jonma\freemocap_data\session_10-15-2022-09_50_10\output_data\post_processed_data\mediaPipeSkel_3d_origin_aligned.npy"
    )
    output_data_folder_path = (
        r"C:\Users\jonma\freemocap_data\session_10-15-2022-09_50_10\output_data"
    )

    convert_mediapipe_npy_to_csv(
        mediapipe_3d_frame_trackedPoint_xyz, output_data_folder_path
    )
